# Rekall Memory Forensics
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Author: Michael Cohen scudette@google.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""IO Abstraction for Rekall.

Since Rekall is a library it should never directly access files: it may be
running on an environment which has no filesystem access for example, or the
files may be stored in an unusual way.

In order to ensure that the file storage mechanism does not need to be hardcoded
in each module, Rekall has an abstracted filesystem access mechanism implemented
through the IO Manager.

The session object should contain an instance of the IOManager() class at the
io_manager attribute, which will be used to create new files, or read from
existing files.
"""

__author__ = "Michael Cohen <scudette@google.com>"

import StringIO
import json
import os
import urllib2
import urlparse
import zipfile

from rekall import registry


class IOManagerError(IOError):
    """An IOError from the IO Manager."""


class IOManager(object):
    """The baseclass for abstracted IO implementations."""

    __metaclass__ = registry.MetaclassRegistry
    __abstract = True

    order = 100

    def __init__(self, urn=None, mode="r"):
        self.mode = mode
        self.urn = urn

    def ListFiles(self):
        """Returns a generator over all the files in this container."""
        return []

    def Create(self, name):
        """Creates a new file in the container.

        Returns a file like object which should support the context manager
        protocol. If the file already exists in the container, overwrite it.

        For example:

        with self.session.io_manager.Create("foobar") as fd:
           fd.Write("hello world")

        Args:
          name: The name of the new file.
        """

    def Open(self, name):
        """Opens a container member for reading.

        This should return a file like object which provides read access to
        container members.

        Raises:
          IOManagerError: If the file is not found.
        """

    def OpenSubContainer(self, name):
        """Opens member as a subcontainer.

        Args:
          name: The name of the subcontainer.

        Returns:
          an instance of IOManager if successful.

        Raises:
          IOManagerError if member is not found or can not be opened as a
          container..
        """

    def GetData(self, name):
        """Get the data object stored at container member.

        This returns an arbitrary python object which is stored in the named
        container member. For example, normally a dict or list. This function
        wraps the Open() method above and add deserialization to retrieve the
        actual object.

        Returns None if the file is not found.
        """
        try:
            return json.load(self.Open(name))
        except IOManagerError:
            return None

    def StoreData(self, name, data, **options):
        """Stores the data in the named container member.

        This serializes the data and stores it in the named member. Not all
        types of data are serializable, so this may raise. For example, when
        using JSON to store the data, arbitrary python objects may not be used.
        """
        with self.Create(name) as fd:
            fd.write(json.dumps(data, sort_keys=True, **options))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class DirectoryIOManager(IOManager):
    """An IOManager which stores everything in files."""

    def __init__(self, output_directory=None, **kwargs):
        super(DirectoryIOManager, self).__init__(**kwargs)

        self.dump_dir = os.path.normpath(os.path.abspath(output_directory))
        self.check_dump_dir(self.dump_dir)
        self.canonical_name = os.path.basename(self.dump_dir)

    def check_dump_dir(self, dump_dir=None):
        if not dump_dir:
            raise IOManagerError("Please specify a dump directory.")

        if self.mode == "w":
            self.EnsureDirectoryExists(self.dump_dir)

        if not os.path.isdir(dump_dir):
            raise IOManagerError("%s is not a directory" % self.dump_dir)

    def _GetAbsolutePathName(self, name):
        path = os.path.normpath(os.path.join(self.dump_dir, name))
        if not path.startswith(self.dump_dir):
            raise IOManagerError("Path name is outside container.")

        return path

    def EnsureDirectoryExists(self, dirname):
        try:
            os.makedirs(dirname)
        except OSError:
            pass

    def ListFiles(self):
        for root, _, files in os.walk(self.dump_dir):
            for f in files:
                path = os.path.normpath(os.path.join(root, f))

                # Return paths relative to the dump dir.
                yield path[len(self.dump_dir) + 1:]

    def Create(self, name):
        path = self._GetAbsolutePathName(name)
        self.EnsureDirectoryExists(os.path.dirname(path))
        return open(path, "wb")

    def Open(self, name):
        path = self._GetAbsolutePathName(name)
        return open(path, "rb")

    def OpenSubContainer(self, name):
        path = self._GetAbsolutePathName(name)

        return Factory(path, mode=self.mode)

    def __str__(self):
        return "Directory:%s" % self.dump_dir


# pylint: disable=protected-access

class SelfClosingFile(StringIO.StringIO):
    def __init__(self, name, manager):
        self.name = name
        self.manager = manager
        StringIO.StringIO.__init__(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.Close()
        else:
            self.manager._Cancel(self.name)

    def Close(self):
        self.manager._Write(self.name, self.getvalue())


class ZipFileManager(IOManager):
    """An IO Manager which stores files in a zip archive."""

    order = 50

    def __init__(self, file_name=None, fd=None, **kwargs):
        super(ZipFileManager, self).__init__(**kwargs)
        if fd is None and not file_name.lower().endswith("zip"):
            if self.mode == "w":
                raise IOManagerError(
                    "Zip files must have the .zip extensions.")
            else:
                # For reading we assume the a name without the .zip extension
                # should have it added.
                file_name = "%s.zip" % file_name

        self.fd = fd
        self.file_name = os.path.normpath(os.path.abspath(file_name))
        self.canonical_name = os.path.splitext(os.path.basename(file_name))[0]
        self._OpenZipFile()

        # The set of outstanding writers. When all outstanding writers have been
        # closed we can flush the ZipFile.
        self._outstanding_writers = set()

    def _OpenZipFile(self, mode=None):
        try:
            if self.fd is None:
                self.zip = zipfile.ZipFile(
                    self.file_name, mode=mode or self.mode,
                    compression=zipfile.ZIP_DEFLATED)

            elif self.mode == "r":
                self.zip = zipfile.ZipFile(self.fd, mode="r")

            else:
                raise IOManagerError("Need a zip filename to write.")

        except zipfile.BadZipfile:
            raise IOManagerError("Unable to read zipfile.")

    def ListFiles(self):
        return self.zip.namelist()

    def _Cancel(self, name):
        self._outstanding_writers.remove(name)

    def _Write(self, name, data):
        self.zip.writestr(name, data)
        self._outstanding_writers.remove(name)
        if not self._outstanding_writers:
            self.zip.close()

            # Reopen the zip file so we may add new members.
            self._OpenZipFile(mode="a")

    def Create(self, name):
        if self.mode != "w":
            raise IOManagerError("Container not opened for writing.")

        result = SelfClosingFile(name, self)
        self._outstanding_writers.add(name)
        return result

    def Open(self, name):
        if self.mode != "r":
            raise IOManagerError("Container not opened for reading.")
        if self.zip is None:
            self._OpenZipFile()

        try:
            return self.zip.open(name)
        except KeyError as e:
            raise IOManagerError(e)

    def OpenSubContainer(self, name):
        if self.mode != "r":
            raise IOManagerError("Only supports nesting for reading.")

        fd = self.Open(name)

        # Currently only support nested zip files here.
        return self.__class__(fd=fd, mode="r")

    def __enter__(self):
        self._outstanding_writers.add(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._outstanding_writers.remove(self)
        if exc_type is None and not self._outstanding_writers:
            self.zip.close()
            if self.mode == "w":
                self._OpenZipFile(mode="a")

    def __str__(self):
        return "ZipFile:%s" % self.file_name


class BuiltInManager(IOManager):
    """An IO manager which uses a python dict."""

    order = 10

    __abstract = True

    # This should contain the data.
    data = {}

    def __init__(self, urn=None, data=None, **kwargs):
        super(BuiltInManager, self).__init__(**kwargs)
        if urn is not None:
            raise IOManagerError("urn specified.")

        if self.mode != "r":
            raise IOManagerError(
                "BuiltInManager can only be opened for reading.")

        if data:
            self.data = data

    def ListFiles(self):
        return self.data.keys()

    def _Cancel(self, name):
        pass

    def _Write(self, name, value):
        self.data[name] = value

    def Create(self, name):
        return SelfClosingFile(name, self)

    def GetData(self, name):
        result = self.data.get(name)
        if not result:
            raise IOManagerError("%s not found." % name)

        return result

    def OpenSubContainer(self, name):
        if name not in self.data:
            raise IOManagerError("%s not found." % name)

        return BuiltInManager(data=self.data[name])

    def __str__(self):
        return "BuiltIn:%s" % self.__class__.__name__


class URLManager(IOManager):
    """Supports openning containers from the web.

    Currenlty we only support openning a zip file fetched from a URL.
    """

    def __init__(self, urn=None, mode="r"):
        super(URLManager, self).__init__(urn=urn, mode=mode)
        if mode != "r":
            raise IOManagerError("%s supports only reading." %
                                 self.__class__.__name__)

        self.url = urlparse.urlparse(urn)
        if self.url.scheme not in ("http", "https"):
            raise IOManagerError("%s supports only http protocol." %
                                 self.__class__.__name__)

    def Create(self, _):
        raise IOManagerError("Write support to http is not supported.")

    def _GetURL(self, name):
        url = self.url._replace(path="%s/%s" % (self.url.path, name))
        return urlparse.urlunparse(url)

    def Open(self, name):
        return urllib2.urlopen(self._GetURL(name))

    def _OpenSubContainer(self, name):
        data = None

        # This can either be a subdirectory
        try:
            # This checks for a 404 Error. Sometimes, a web server can return
            # the zip file here, even though there might not be a zip extension.
            url = self._GetURL(name)
            data = urllib2.urlopen(self._GetURL(name)).read()

            try:
                return ZipFileManager(fd=StringIO.StringIO(data),
                                      file_name=url)
            except IOError:
                return URLManager(url)

        except IOError as e:
            raise IOManagerError(e)

    def OpenSubContainer(self, name):
        # Or it can be a zip file.
        try:
            return self._OpenSubContainer(name)
        except IOError:
            return self._OpenSubContainer(name + ".zip")

    def __str__(self):
        return "URL:%s" % self.urn


def Factory(urn, mode="r", **kwargs):
    """Try to instantiate the IOManager class."""
    for cls in sorted(IOManager.classes.values(), key=lambda x: x.order):
        try:
            return cls(urn, mode=mode, **kwargs)
        except IOError:
            pass

    raise IOManagerError(
        "Unable to find any managers which can work on %s" % urn)



# Rekall Memory Forensics
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

__author__ = "Michael Cohen <scudette@gmail.com>"

"""Implement OSX support."""


from rekall import obj
from rekall.plugins.overlays import basic

darwin_overlay = {

    }


class Darwin32(basic.Profile32Bits, basic.BasicWindowsClasses):
    """A Darwin profile."""
    _md_os = "darwin"
    _md_memory_model = "32bit"
    _md_type = "Kernel"

    def __init__(self, **kwargs):
        super(Darwin32, self).__init__(**kwargs)
        self.add_classes(dict(
                ))
        self.add_overlay(darwin_overlay)
        self.add_constants(default_text_encoding="utf8")

class Darwin64(basic.ProfileLP64, Darwin32):
    """Support for 64 bit darwin systems."""



# Load the core modules
# pylint: disable=unused-import

from rekall import utils

from rekall.plugins.addrspaces import amd64
from rekall.plugins.addrspaces import crash
from rekall.plugins.addrspaces import hibernate
from rekall.plugins.addrspaces import intel
from rekall.plugins.addrspaces import macho
from rekall.plugins.addrspaces import mmap_address_space
from rekall.plugins.addrspaces import standard
from rekall.plugins.addrspaces import vboxelf

utils.ConditionalImport("rekall.plugins.addrspaces.accelerated")
utils.ConditionalImport("rekall.plugins.addrspaces.ewf")

# If we are running on windows, load the windows specific AS.
try:
    import rekall.plugins.addrspaces.win32
except ImportError:
    pass



# Rekall Memory Forensics
# Copyright (C) 2012 Michael Cohen
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
"""This is a windows specific address space."""
import struct
import win32file

from rekall import addrspace
from rekall import utils


def CTL_CODE(DeviceType, Function, Method, Access):
    return (DeviceType<<16) | (Access << 14) | (Function << 2) | Method


# IOCTLS for interacting with the driver.
INFO_IOCTRL = CTL_CODE(0x22, 0x103, 0, 3)

PAGE_SHIFT = 12


class Win32FileAddressSpace(addrspace.RunBasedAddressSpace):
    """ This is a direct file AS for use in windows.

    In windows, in order to open raw devices we need to use the win32 apis. This
    address space allows us to open the raw device as exported by e.g. the
    winpmem driver.
    """

    __name = "win32file"

    ## We should be the AS of last resort but in front of the non win32 version.
    order = 90
    PAGE_SIZE = 0x10000
    _md_image = True

    def __init__(self, filename=None, **kwargs):
        super(Win32FileAddressSpace, self).__init__(**kwargs)

        self.as_assert(self.base == None, 'Must be first Address Space')

        path = filename or (self.session and self.session.filename)
        self.as_assert(path, "Filename must be specified in session (e.g. "
                       "session.filename = 'MyFile.raw').")

        self.fname = path
        self.fhandle = win32file.CreateFile(
            path,
            win32file.GENERIC_READ | win32file.GENERIC_WRITE,
            win32file.FILE_SHARE_READ | win32file.FILE_SHARE_WRITE,
            None,
            win32file.OPEN_EXISTING,
            win32file.FILE_ATTRIBUTE_NORMAL,
            None)

        # Try to get the memory runs from the winpmem driver.
        self.runs = []
        try:
            self.ParseMemoryRuns()
        except Exception:
            if self.session.debug:
                import pdb
                pdb.post_mortem()

            self.runs = [[0, 0, win32file.GetFileSize(self.fhandle)]]

        # IO on windows is extremely slow so we are better off using a
        # cache.
        self.cache = utils.FastStore(1000)

    FIELDS = (["CR3", "NtBuildNumber", "KernBase", "KDBG"] +
              ["KPCR%02d" % i for i in range(32)] +
              ["PfnDataBase", "PsLoadedModuleList", "PsActiveProcessHead"] +
              ["Padding%s" % i for i in range(0xff)] +
              ["NumberOfRuns"])

    def ParseMemoryRuns(self):
        result = win32file.DeviceIoControl(
            self.fhandle, INFO_IOCTRL, "", 102400, None)

        fmt_string = "Q" * len(self.FIELDS)
        self.memory_parameters = dict(zip(self.FIELDS, struct.unpack_from(
                    fmt_string, result)))

        self.dtb = self.memory_parameters["CR3"]
        self.session.StoreParameter("dtb", int(self.dtb))

        self.kdbg = self.memory_parameters["KDBG"]
        self.session.StoreParameter("kdbg", int(self.kdbg))

        offset = struct.calcsize(fmt_string)

        for x in range(self.memory_parameters["NumberOfRuns"]):
            start, length = struct.unpack_from("QQ", result, x * 16 + offset)
            self.runs.append((start, start, length))

    def _read_chunk(self, addr, length):
        offset, available_length = self._get_available_buffer(addr, length)
        if offset is None:
            return "\x00" * min(length, available_length)

        win32file.SetFilePointer(self.fhandle, offset, 0)
        _, data = win32file.ReadFile(
            self.fhandle, min(length, available_length))

        return data

    def write(self, addr, data):
        length = len(data)
        offset, available_length = self._get_available_buffer(addr, length)
        if offset is None:
            return

        to_write = min(len(data), available_length)
        win32file.SetFilePointer(self.fhandle, offset, 0)

        win32file.WriteFile(self.fhandle, data[:to_write])

        return to_write

    def close(self):
        win32file.CloseHandle(self.fhandle)



# Rekall Memory Forensics
# Copyright (C) 2012 Michael Cohen
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This module implements a text based render.

A renderer is used by plugins to produce formatted output.
"""
try:
    import curses
except ImportError:
    curses = None

import logging
import json
import re
import os
import string
import subprocess
import sys
import tempfile
import textwrap
import time

from rekall import config
from rekall import obj
from rekall import utils
from rekall import registry
from rekall import constants


config.DeclareOption(
    "--pager", default=os.environ.get("PAGER"), group="Interface",
    help="The pager to use when output is larger than a screen full.")


config.DeclareOption(
    "--renderer", default="TextRenderer", group="Interface",
    help="The renderer to use. e.g. (TextRenderer, "
    "JsonRenderer).")

config.DeclareOption(
    "--nocolors", default=False, action="store_true", group="Interface",
    help="If set suppress outputting colors.")

config.DeclareOption(
    "--logging", default="error", choices=[
        "debug", "info", "warning", "critical", "error"],
    help="Logging level to show messages.")

config.DeclareOption(
    "-v", "--verbose", default=False, action="store_true",
    help="Set logging to debug level.")

config.DeclareOption(
    "--debug", default=False, action="store_true",
    help="If set we break into the debugger on error conditions.")


HIGHLIGHT_SCHEME = dict(
    important=("WHITE", "RED"),
    good=("GREEN", None),
    neutral=(None, None),
    )

class Pager(object):
    """A wrapper around a pager.

    The pager can be specified by the session. (eg. session.state.pager =
    'less') or in an PAGER environment var.
    """
    # Default encoding is utf8
    encoding = "utf8"

    def __init__(self, session=None, encoding=None):
        # More is the least common denominator of pagers :-(. Less is better,
        # but most is best!
        self.pager_command = session.state.pager or os.environ.get("PAGER")
        self.encoding = encoding or session.encoding or sys.stdout.encoding

        # Make a temporary filename to store output in.
        self.fd, self.filename = tempfile.mkstemp(prefix="rekall")

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        # Delete the temp file.
        try:
            os.unlink(self.filename)
        except OSError:
            pass

    def write(self, data):
        # Encode the data according to the output encoding.
        data = utils.SmartUnicode(data).encode(self.encoding, "replace")
        try:
            if sys.platform in ["win32"]:
                data = data.replace("\n", "\r\n")

            os.write(self.fd, data)
        # This can happen if the pager disappears in the middle of the write.
        except IOError:
            pass

    def flush(self):
        """Wait for the pager to be exited."""
        os.close(self.fd)

        try:
            args = dict(filename=self.filename)
            # Allow the user to interpolate the filename in a special way,
            # otherwise just append to the end of the command.
            if "%" in self.pager_command:
                pager_command = self.pager_command % args
            else:
                pager_command = self.pager_command + " %s" % self.filename

            subprocess.call(pager_command, shell=True)

        # Allow the user to break out from waiting for the command.
        except KeyboardInterrupt:
            pass
        finally:
            try:
                os.unlink(self.filename)
            except OSError:
                pass


class UnicodeWrapper(object):
    """A wrapper around a file like object which guarantees writes in utf8."""

    def __init__(self, fd, encoding='utf8'):
        self.fd = fd
        self.encoding = encoding

    def write(self, data):
        data = utils.SmartUnicode(data).encode(self.encoding, "replace")
        self.fd.write(data)

    def flush(self):
        self.fd.flush()

    def isatty(self):
        return self.fd.isatty()


class Formatter(string.Formatter):
    """A formatter which supports extended formating specs."""
    # This comes from http://docs.python.org/library/string.html
    # 7.1.3.1. Format Specification Mini-Language
    standard_format_specifier_re = re.compile("""
(?P<fill>[^{}<>=^bcdeEfFgGnLosxX])?   # The fill parameter. This can not be a
                                     # format string or it is ambiguous.
(?P<align>[<>=^])?     # The alignment.
(?P<sign>[+\- ])?      # Sign extension.
(?P<hash>\#)?          # Hash means to preceed the whole thing with 0x.
(?P<zerofill>0)?       # Should numbers be zero filled.
(?P<width>\d+)?        # The minimum width.
(?P<comma>,)?
(?P<precision>.\d+)?   # Precision
(?P<type>[bcdeEfFgGnosxXL%])?  # The format string (Not all are supported).
""", re.X)

    def format_field(self, value, format_spec):
        """Format the value using the format_spec.

        The aim of this function is to remove the delegation to __format__() on
        the object. For our needs we do not want the object to be responsible
        for its own formatting since it is not aware of the renderer itself.

        A rekall.obj.BaseObject instance must support the following
        formatting operations:

        __unicode__
        __str__
        __repr__
        and may also support __int__ (for formatting in hex).
        """
        m = self.standard_format_specifier_re.match(format_spec)
        if not m:
            raise re.error("Invalid regex")

        fields = m.groupdict()

        # Format the value according to the basic type.
        type = fields["type"] or "s"
        try:
            value = getattr(
                self, "format_type_%s" % type)(value, fields)
        except AttributeError:
            raise re.error("No formatter for type %s" % type)

        try:
            return format(value, format_spec)
        except ValueError:
            return str(value)

    def format_type_s(self, value, fields):
        try:
            # This is required to allow BaseObject to pass non unicode returns
            # from __unicode__ (e.g. NoneObject).
            result = value.__unicode__()
        except AttributeError:
            result = utils.SmartUnicode(value)

        # None objects get a -.
        if result is None or isinstance(result, obj.NoneObject):
            return "-" * int(fields['width'] or "1")

        return result

    def format_type_x(self, value, fields):
        return int(value)

    def format_type_X(self, value, fields):
        return int(value)

    def format_type_r(self, value, fields):
        return repr(value)

    def format_type_f(self, value, fields):
        if isinstance(value, (float, int, long)):
            return float(value)

        return value

    def format_type_L(self, value, fields):
        """Support extended list format."""
        return ", ".join(value)


class TextColumn(object):
    """An implementation of a Column."""

    def __init__(self, name=None, cname=None, formatstring="s", address_size=14,
                 header_format=None, table=None):
        self.name = name or "-"
        self.cname = cname or "-"
        self.table = table
        self.wrap = None

        # How many places should addresses be padded?
        self.address_size = address_size
        self.parse_format(formatstring=formatstring,
                          header_format=header_format)

        # The format specifications is a dict.
        self.formatter = Formatter()
        self.header_width = 0

    def parse_format(self, formatstring=None, header_format=None):
        """Parse the format string into the format specification.

        We support some extended forms of format string which we process
        especially here:

        [addrpad] - This is a padded address to width self.address_size.
        [addr] - This is a non padded address.
        [wrap:width] - This wraps a stringified version of the target in the
           cell.
        """
        # Leading ! turns off eliding.
        if formatstring.startswith("!"):
            self.table.elide = True
            formatstring = formatstring[1:]

        # This means unlimited width.
        if formatstring == "":
            self.header_format = self.formatstring = ""

            # Eliding is not possible without column width limits.
            self.table.elide = False
            return

        m = re.search("\[addrpad\]", formatstring)
        if m:
            self.formatstring = "#0%sx" % self.address_size
            self.header_format = "^%ss" % self.address_size
            # Never elide addresses - makes them unreadable.
            self.table.elide = False
            return

        m = re.search("\[addr\]", formatstring)
        if m:
            self.formatstring = ">#%sx" % self.address_size
            self.header_format = "^%ss" % self.address_size
            self.table.elide = False
            return

        # Look for the wrap specifier.
        m = re.search("\[wrap:([^\]]+)\]", formatstring)
        if m:
            self.formatstring = "s"
            self.wrap = int(m.group(1))
            self.header_format = "<%ss" % self.wrap
            return

        # Fall through to a simple format specifier.
        self.formatstring = formatstring

        if header_format is None:
            self.header_format = re.sub("[Xx]", "s", formatstring)

    def render_header(self):
        """Renders the cell header."""
        header_cell = self.render_cell(
            self.name, formatstring=self.header_format, elide=False)
        self.header_width = max([len(line) for line in header_cell])

        # Append a dashed line as a table header separator.
        header_cell.append("-" * self.header_width)

        return header_cell

    def elide_string(self, string, length):
        """Adds three dots in the middle of a string if it is longer than length"""
        if length == -1:
            return string

        if len(string) < length:
            return (" " * (length - len(string))) + string

        elif len(string) == length:
            return string

        else:
            if length < 5:
                logging.error("Cannot elide a string to length less than 5")

            even = ((length + 1) % 2)
            length = (length - 3) / 2
            return string[:length + even] + "..." + string[-length:]

    def render_cell(self, target, formatstring=None, elide=None):
        """Renders obj according to the format string."""
        if formatstring is None:
            formatstring = self.formatstring

        if isinstance(target, Colorizer):
            result = []
            for x in self.render_cell(target.target, formatstring=formatstring,
                                      elide=elide):
                result.append(target.Render(x))
            return result

        # For NoneObjects we just render dashes. (Other renderers might want to
        # actually record the error, we ignore it here.).
        elif target is None or isinstance(target, obj.NoneObject):
            return ['-' * len(self.formatter.format_field(1, formatstring))]

        # Simple formatting.
        result = self.formatter.format_field(target, formatstring).splitlines()

        # Support line wrapping.
        if self.wrap:
            old_result = result
            result = []
            for line in old_result:
                result.extend(textwrap.wrap(line, self.wrap))

        elif elide is None:
            elide = self.table.elide

        if elide:
            # we take the header width as the maximum width of this column.
            result = [
                self.elide_string(line, self.header_width) for line in result]

        if isinstance(target, bool):
            color = "GREEN" if target else "RED"
            result = [self.table.renderer.color(x, foreground=color) for x in result]

        return result or [""]


class TextTable(object):
    """A table is a collection of columns.

    This table formats all its cells using proportional text font.
    """

    def __init__(self, columns=None, tablesep=" ", elide=False,
                 suppress_headers=False, address_size=10, renderer=None):
        self.columns = [
            TextColumn(*args, address_size=address_size, table=self)
            for args in columns]

        self.renderer = renderer  # Our parent renderer.
        self.tablesep = tablesep
        self.elide = elide
        self.suppress_headers = suppress_headers

    def write_row(self, renderer, cells, highlight=False):
        """Writes a row of the table.

        Args:
          renderer: The renderer we use to write on.
          cells: A list of cell contents. Each cell content is a list of lines
            in the cell.
        """
        foreground, background = HIGHLIGHT_SCHEME.get(
            highlight, (None, None))

        # Ensure that all the cells are the same width.
        justified_cells = []
        cell_widths = []
        max_height = 0
        for cell in cells:
            max_width = max([len(line) for line in cell])
            max_height = max(max_height, len(cell))
            justified_cell = []
            for line in cell:
                justified_cell.append(line + (' ' * (max_width-len(line))))

            justified_cells.append(justified_cell)
            cell_widths.append(max_width)

        for line in range(max_height):
            line_components = []
            for i in range(len(justified_cells)):
                try:
                    line_components.append(justified_cells[i][line])
                except IndexError:
                    line_components.append(" " * cell_widths[i])

            renderer.write(
                renderer.color(
                    self.tablesep.join(line_components),
                    foreground=foreground, background=background) + "\n")

    def render_header(self, renderer):
        # The headers must always be calculated so we can work out the column
        # widths.
        headers = [c.render_header() for c in self.columns]

        if not self.suppress_headers:
            self.write_row(renderer, headers)

    def render_row(self, renderer, row=None, highlight=None):
        self.write_row(
            renderer,
            [c.render_cell(obj) for c, obj in zip(self.columns, row)],
            highlight=highlight)



class RendererBaseClass(object):
    """All renderers inherit from this."""

    __metaclass__ = registry.MetaclassRegistry

    def __init__(self, session=None, fd=None, paging_limit=50):
        self.session = session
        self.paging_limit = paging_limit
        self.fd = fd
        self.isatty = False
        self.formatter = Formatter()
        self.colorizer = Colorizer(
            fd, nocolor=session.GetParameter("nocolors") if session else False)

    def start(self, plugin_name=None, kwargs=None):
        """The method is called when new output is required.

        Metadata about the running plugin is provided so the renderer may log it
        if desired.

        Args:
           plugin_name: The name of the plugin which is running.
           kwargs: The args for this plugin.
        """

    def end(self):
        """Tells the renderer that we finished using it for a while."""

    def write(self, data):
        """Renderer should write some data."""

    def section(self, name=None):
        """Start a new section.

        Sections are used to separate distinct entries (e.g. reports of different files).
        """
        self.write("*" * 50 + "\n")

    def format(self, formatstring, *data):
        """Write formatted data.

        For renderers that need access to the raw data (e.g. to check for
        NoneObjects), it is preferred to call this method directly rather than
        to format the string in the plugin itself.

        By default we just call the format string directly.
        """
        self.write(self.formatter.format(formatstring, *data))

    def flush(self):
        """Renderer should flush data."""

    def table_header(self, title_format_list = None, suppress_headers=False,
                     name=None):
        """Table header renders the title row of a table.

        This also stores the header types to ensure everything is formatted
        appropriately.  It must be a list of tuples rather than a dict for
        ordering purposes.

        Args:
           title_format_list: A list of (Name, formatstring) tuples describing
              the table headers.

           suppress_headers: If True table headers will not be written (still
              useful for formatting).

           name: The name of this table.
        """

    def table_row(self, *args, **kwargs):
        """Outputs a single row of a table.

        Supported kwargs:
          highlight: Highlight this raw according to the color scheme
            (e.g. important, good)
        """

    def record(self, record_data):
        """Writes a single complete record.

        A record consists of one object of related fields.

        Args:
          data: A list of tuples (name, short_name, formatstring, data)
        """
        for name, short_name, formatstring, data in record_data:
            self.format("%s: %s\n" % (name, formatstring), data)

        self.format("\n")

    def color(self, target, **kwargs):
        return self.colorizer.Render(target, **kwargs)


class TextRenderer(RendererBaseClass):
    """Plugins can receive a renderer object to assist formatting of output."""

    tablesep = " "
    elide = False
    spinner = "/-\|"
    last_spin_time = 0
    last_spin = 0
    last_message_len = 0
    isatty = False

    def __init__(self, tablesep=" ", elide=False, max_data=1024*1024,
                 paging_limit=None, **kwargs):
        super(TextRenderer, self).__init__(**kwargs)
        self.tablesep = tablesep
        self.elide = elide
        self.paging_limit = paging_limit

        # We keep the data that we produce in memory for while.
        self.data = []
        self.max_data = max_data

        # Make sure that our output is unicode safe.
        self.fd = UnicodeWrapper(self.fd or sys.stdout)

    def start(self, plugin_name=None, kwargs=None):
        """The method is called when new output is required.

        Args:
           plugin_name: The name of the plugin which is running.
           kwargs: The args for this plugin.
        """
        # This handles the progress messages from rekall for the duration of
        # the rendering but only if we output to a tty.
        if self.session and self.fd.isatty():
            self.session.progress = self.RenderProgress

    def end(self):
        """Tells the renderer that we finished using it for a while."""
        # Remove the progress handler from the session.
        if self.session:
            self.session.progress = None

    def format(self, formatstring, *data):
        self.ClearProgress()
        super(TextRenderer, self).format(formatstring, *data)

    def write(self, data):
        self.data.append(data)

        if self.paging_limit is None:
            self.fd.write(data)
            return

        elif len(self.data) < self.paging_limit:
            self.fd.write(data)
            self.fd.flush()

        elif len(self.data) == self.paging_limit:
            self.fd.write(
                self.color("Please wait while the rest is paged...",
                           foreground="YELLOW") + "\r\n")
            self.fd.flush()

    def flush(self):
        self.data = []
        self.ClearProgress()
        self.fd.flush()

    def table_header(self, columns = None, suppress_headers=False,
                     **kwargs):
        """Table header renders the title row of a table.

        This also stores the header types to ensure everything is formatted
        appropriately.  It must be a list of tuples rather than a dict for
        ordering purposes.

        Args:
           columns: A list of (Name, formatstring) tuples describing
              the table columns.

           suppress_headers: If True table headers will not be written (still
              useful for formatting).
        """
        # Determine the address size
        address_size = 14
        if (self.session and self.session.profile and
            self.session.profile.metadata("memory_model") == "32bit"):
            address_size = 10

        self.table = TextTable(columns=columns, tablesep=self.tablesep,
                               suppress_headers=suppress_headers,
                               address_size=address_size,
                               renderer=self)
        self.table.render_header(self)

    def table_row(self, *args, **kwargs):
        """Outputs a single row of a table"""
        return self.table.render_row(self, row=args, **kwargs)

    def ClearProgress(self):
        """Delete the last progress message."""
        if not sys.stdout.isatty():
            return

        # Wipe the last message.
        sys.stdout.write("\r" + " " * self.last_message_len + "\r")
        sys.stdout.flush()

    def RenderProgress(self, message="%(spinner)s", *args, **kwargs):
        if not sys.stdout.isatty():
            return

        # Only write once per second.
        now = time.time()
        force = kwargs.get("force")

        if force or now > self.last_spin_time + 0.2:
            self.last_spin_time = now
            self.last_spin += 1

            # Only expand variables when we need to.
            if kwargs:
                kwargs["spinner"] = self.spinner[
                    self.last_spin % len(self.spinner)]

                message = message % kwargs
            elif args:
                message = message % args

            self.ClearProgress()

            sys.stdout.write(message + "\r")
            self.last_message_len = len(message)
            sys.stdout.flush()


class JsonFormatter(Formatter):
    """A formatter for json object."""

    def format_dict(self, value):
        result = []
        for k, v in value.items():
            result.append((k, self.format_field(v, "s")))

        return dict(result)

    def format_field(self, value, format_spec):
        """The json formatter aims to capture as many properties of the value as
        possible.
        """
        # We try to capture as much information about this object. Hopefully
        # this should be enough to reconstruct this object later.
        if isinstance(value, obj.BaseObject):
            result = dict(rekall_type=value.obj_type,
                          rekall_name=value.obj_name,
                          rekall_offset=value.obj_offset,
                          rekall_vm=str(value.obj_vm))

            for method in ["__unicode__", "__int__", "__str__"]:
                try:
                    result['value'] = self.format_field(
                        getattr(value, method)(), "s")['value']
                    break
                except (AttributeError, ValueError):
                    pass


            return result

        # If it is a simple type, just pass it as is.
        if isinstance(value, (int, long, basestring)):
            return dict(value=value)

        # If it is a NoneObject dump out the error
        if isinstance(value, obj.NoneObject):
            return dict(rekall_type=value.__class__.__name__,
                        rekall_reason=value.reason,
                        value=None)

        # Fall back to just formatting it.
        return super(JsonFormatter, self).format_field(value, format_spec)


class JsonColumn(TextColumn):
    """A column in a json table."""
    def __init__(self, name=None, cname=None, format_spec=None, **kwargs):
        self.formatter = JsonFormatter()
        self.name = name
        self.cname = cname

    def render_header(self):
        return self.cname

    def render_cell(self, target):
        return self.formatter.format_field(target, "s")


class JsonTable(TextTable):
    def __init__(self, columns=None, **kwargs):
        self.columns = [JsonColumn(*args) for args in columns]

    def render_header(self, renderer):
        renderer.table_data['headers'] = [c.render_header() for c in self.columns]

    def get_header(self, renderer):
        return [c.render_header() for c in self.columns]

    def render_row(self, renderer, row=None, **_):
        data = {}
        for c, obj in zip(self.columns, row):
            data[c.cname] = c.render_cell(obj)
        renderer.table_data.append(data)


class JsonRenderer(TextRenderer):
    """Render the output as a json object."""

    def start(self, plugin_name=None, kwargs=None):
        self.formatter = JsonFormatter()

        # We store the data here.
        self.data = dict(plugin_name=plugin_name,
                         tool_name="rekall-ng",
                         tool_version=constants.VERSION,
                         kwargs=self.formatter.format_dict(kwargs or {}),
                         data=[])

        super(JsonRenderer, self).start(plugin_name=plugin_name,
                                        kwargs=kwargs)
        self.headers = []

    def end(self):
        # Just dump out the json object.
        self.fd.write(json.dumps(self.data, indent=4))

    def format(self, formatstring, *args):
        statement = [formatstring]
        for arg in args:
            # Just store the statement in the output.
            statement.append(self.formatter.format_field(arg, "s"))

        self.data['data'].append(statement)

    def table_header(self, columns = None, **kwargs):
        self.table = JsonTable(columns=columns)

        # This is the current table - the JsonTable object will write on it.
        self.table_data = []

        # Append it to the data.
        self.data['data'] = self.table_data

        # Write the headers.
        self.headers = self.table.get_header(self)

    def write(self, data):
        self.data['data'].append(data)


class TestRenderer(TextRenderer):
    """A special renderer which makes parsing the output of tables easier."""

    def __init__(self, **kwargs):
        super(TestRenderer, self).__init__(tablesep="||", **kwargs)


class Colorizer(object):
    """An object which makes its target colorful."""

    COLORS = "BLACK BLUE GREEN CYAN RED MAGENTA YELLOW WHITE"
    COLOR_MAP = dict([(x, i) for i, x in enumerate(COLORS.split())])

    terminal_capable = False

    def __init__(self, stream, nocolor=False):
        """Initialize a colorizer.

        Args:
          stream: The stream to write to.

          nocolor: If True we suppress using colors, even if the output stream
             can support them.
        """
        if stream is None:
            stream = sys.stdout

        if nocolor:
            self.terminal_capable = False
            return

        try:
            if curses and stream.isatty():
                curses.setupterm()
                self.terminal_capable = True
        except AttributeError:
            pass

    def tparm(self, capabilities, *args):
        """A simplified version of tigetstr without terminal delays."""
        for capability in capabilities:
            term_string = curses.tigetstr(capability)
            if term_string is not None:
                term_string = re.sub("\$\<[^>]+>", "",  term_string)
                break

        try:
            return curses.tparm(term_string, *args)
        except Exception, e:
            logging.debug("Unable to set tparm: %s" % e)
            return ""

    def Render(self, string, foreground=None, background=None):
        """Decorate the string with the ansii escapes for the color."""
        if (not self.terminal_capable or
            foreground not in self.COLOR_MAP or
            foreground not in self.COLOR_MAP):
            return utils.SmartUnicode(string)

        escape_seq = ""
        if background:
            escape_seq += self.tparm(["setb", "setab"], self.COLOR_MAP[background])

        if foreground:
            escape_seq += self.tparm(["setf", "setaf"], self.COLOR_MAP[foreground])

        return (escape_seq + utils.SmartUnicode(string) +
                self.tparm(["sgr0"]))




# Rekall Memory Forensics
# Copyright (C) 2012 Michael Cohen <scudette@gmail.com>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""Provides the primitives needed to disassemble code using distorm3."""

# This stuff is just here to make pyinstaller pick up on it. Due to the way
# distorm3 uses ctypes, pyinstaller misses the imports. Note that the following
# patch should also be applied to distorm3/__init__.py to support freezing:
#
#    # Guess the DLL filename and load the library.
#    if getattr(sys, "frozen", None):
#        _distorm_path = '.'
#    else:
#        _distorm_path = split(__file__)[0]
try:
    from ctypes import cdll

    cdll.LoadLibrary("distorm3.dll")
except Exception:
    pass

import distorm3
import re

from rekall import args
from rekall import plugin
from rekall import testlib


class Disassemble(plugin.Command):
    """Disassemble the given address space."""

    __name = "dis"

    @classmethod
    def args(cls, parser):
        super(Disassemble, cls).args(parser)
        parser.add_argument("offset", action=args.IntParser,
                            help="An offset to disassemble.")

        parser.add_argument("-a", "--address_space", default="K",
                            help="The address space to use.")

        parser.add_argument(
            "-l", "--length", default=50,
            help="The number of instructions (lines) to disassemble.")

    def __init__(self, offset=0, address_space=None, length=50, mode=None,
                 suppress_headers=False, target=None, **kwargs):
        """Dumps a disassembly of a location.

        Args:
          address_space: The address_space to read from.
          offset: The offset to read from.
          length: The number of instructions (lines) to disassemble.
          mode: The mode (32/64 bit)- if not set taken from profile.
          suppress_headers: If set we do not write headers.
          target: An ObjBase instance. If specified we do not need the offset
            and address_space.
        """
        super(Disassemble, self).__init__(**kwargs)
        if target is not None:
            address_space = target.obj_vm
            offset = target.offset

        load_as = self.session.plugins.load_as(session=self.session)
        self.address_space = load_as.ResolveAddressSpace(address_space)
        self.offset = offset
        self.length = length
        self.suppress_headers = suppress_headers
        self.mode = mode or self.session.profile.metadata(
            "memory_model", "32bit")

        if self.mode == "32bit":
            self.distorm_mode = distorm3.Decode32Bits
        else:
            self.distorm_mode = distorm3.Decode64Bits

    def disassemble(self, offset):
        """Disassemble the number of instructions required.

        Returns:
          A tuple of (Address, Opcode, Instructions).
        """
        data = self.address_space.read(offset, self.length * 10)
        iterable = distorm3.DecodeGenerator(
            int(offset), data, self.distorm_mode)

        for (offset, _size, instruction, hexdump) in iterable:
            yield offset, hexdump, instruction

    def render(self, renderer):
        """Disassemble code at a given address.

        Disassembles code starting at address for a number of bytes
        given by the length parameter (default: 128).

        Note: This feature requires distorm, available at
            http://www.ragestorm.net/distorm/

        The mode is '32bit' or '64bit'. If not supplied, the disasm
        mode is taken from the profile.
        """
        renderer.table_header([('Address', "cmd_address", '[addrpad]'),
                               ('Op Codes', "opcode", '<20'),
                               ('Instruction', "instruction", '<40'),
                               ('Comment', "comment", "")],
                              suppress_headers=self.suppress_headers)

        offset = 0
        regex = re.compile("0x[0-9a-fA-F]+$")
        for i, (offset, hexdump, instruction) in enumerate(
            self.disassemble(self.offset)):
            if i >= self.length:
                break

            comment = ""
            match = regex.search(instruction)
            if match:
                operand = int(match.group(0), 16)

                # Try to locate the symbol below it.
                symbol, _ = self.address_space.kb.GetSpan(operand)
                if symbol:
                    comment = "%s+0x%X" % (symbol[0].obj_name,
                                           operand - int(symbol[0]))

            renderer.table_row(offset, hexdump, instruction, comment)

        # Continue from where we left off when the user calls us again with the
        # v() plugin.
        self.offset = offset


class TestDisassemble(testlib.SimpleTestCase):
    PARAMETERS = dict(
        commandline="dis %(func)s",
        func=0x805031be
        )



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
"""Enumerate all kernel modules."""

__author__ = "Michael Cohen <scudette@google.com>"
import bisect

from rekall import obj
from rekall.plugins.darwin import common


class DarwinLsmod(common.DarwinPlugin):
    """Lists all kernel modules."""

    __name = "lsmod"

    modlist = None
    mod_lookup = None

    def ResolveSymbolName(self, addr):
        """Resolve a pointer into a name.

        If the symbol name is known we return that, otherwise we try to find the
        containing module, or else we return None of we dont know its name..
        """
        # Try to resolve the address from the profile.
        return (self.profile.get_constant_by_address(addr) or

                # Search for a module which contains this address.
                self.find_module(addr).name)

    def find_module(self, addr):
        """Returns the module which contains this address.

        If the address does not exist in any module, returns a NoneObject.
        """
        if self.modlist is None:
            self._make_cache()

        addr = obj.Pointer.integer_to_address(addr)
        pos = bisect.bisect_right(self.modlist, addr) - 1
        if pos == -1:
            return obj.NoneObject("Unknown address")

        # Get the module.
        module = self.mod_lookup[self.modlist[pos]]
        start = int(module.address)

        if (addr >= start and
            addr < start + module.m("size")):
            return module

        return obj.NoneObject("Unknown address")

    def _make_cache(self):
        self.mod_lookup = {}

        for module in self.get_module_list():
            self.mod_lookup[int(module.address.deref())] = module

        # The start addresses in sorted order.
        self.modlist = sorted(self.mod_lookup.keys())

    def get_module_list(self):
        # The kernel is also included in the module list to make it easier to
        # local pointers inside it.
        # See: xnu-2422.1.72/bsd/dev/dtrace/dtrace.c: 19843
        kernel = self.profile.get_constant_object(
            "_g_kernel_kmod_info", "kmod_info")
        if kernel:
            yield kernel

        module = self.profile.get_constant_object(
            "_kmod",
            target="Pointer",
            target_args=dict(
                target="kmod_info"
                ),
            vm=self.kernel_address_space)

        # walk the modules list
        while module:
            yield module

            module = module.next

    def render(self, renderer):
        renderer.table_header([("Address", "address", "[addrpad]"),
                               ("Size", "size", "[addrpad]"),
                               ("Refs", "refs", ">8"),
                               ("Version", "version", ">12"),
                               ("Name", "name", "")])

        for mod in self.get_module_list():
            renderer.table_row(mod.address,
                               mod.m("size"),
                               mod.reference_count,
                               mod.version,
                               mod.name)



#!/usr/bin/env python2

# Rekall Memory Forensics
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Author: Michael Cohen scudette@google.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
Converts Volatility profile files into the Rekall format.

The Volatility profiles are derived by dumping debugging symbols using various
means into a zip file:

- On Linux the output of dwarfdump is stored and parsed on each execution. The
  constants are just copied from the System map.

- On OSX the symbols are produced using the dsymutil tool while the vtypes are
  python files.

- On Windows the vtypes are python files which must be executed.

Rekall profiles are more structured. All profiles contain a metadata file within
the zip archive called "metadata" which simply contains key value pairs encoded
using json. For example:

{
 # This must point at the implementation of this profile (i.e. the class which
 # should be created). Valid values include Linux32, Linux64, WinXPSP1x86
 # etc. You can use the 'info' plugin to see which classes already exist.

 "ProfileClass": "Linux64"

 # This is the name of a member inside this zip file which contains the
 # constant list.

 "Constants":  "System.map.json"

 # This points at a json file within this zip file which contains the vtype
 # definitions for this profile.

 "VTypes": "vtypes.json"
}

We chose to use json to store the vtype data structures because loading json
files in python is extremely quick and leads to much faster start up times than
having to parse the vtypes in other formats (We do not allow loading of vtypes
inside python files because this may lead to arbitrary code execution since the
vtype file needs to be evaluated.).

Often users already have profiles created for Volatility which they way to use
in Rekall. Rather than fall back to the slow and inefficient parsing of these
profiles, Rekall allows users to convert the old profile into a new, efficient
profile representation. This is what this module does with the convert command.

For example, suppose you have an existing profile created for use in Volatility,
you can just convert it to the rekall format:

./tools/profile_converter.py convert Ubuntu-3.0.0-32-generic-pae.zip \
   Ubuntu-3.0.0-32-generic-pae.rekall.zip

$ ls -l Ubuntu-3.0.0-32-generic-pae.*
-rw-r----- 1 scudette g 643711 Dec 12 02:12 Ubuntu-3.0.0-32-generic-pae.rekall.zip
-rw-r----- 1 scudette g 726480 Dec 12 00:30 Ubuntu-3.0.0-32-generic-pae.zip

Now simply specify the rekall profile using the --profile command line arg.
"""

__author__ = "Michael Cohen <scudette@google.com>"

import logging
import json
import os
import re
import StringIO

from rekall import builtin_profiles
from rekall import io_manager
from rekall import plugin
from rekall import registry

from rekall.plugins.overlays.linux import dwarfdump
from rekall.plugins.overlays.linux import dwarfparser


class ProfileConverter(object):
    """Base class for converters."""

    __metaclass__ = registry.MetaclassRegistry
    __abstract = True

    def __init__(self, input, output, profile_class=None):
        self.input = input
        self.output = output
        self.profile_class = profile_class

    def SelectFile(self, regex):
        """Reads the content of the first file which matches regex."""
        for f in self.input.ListFiles():
            if re.search(regex, f, re.I):
                return self.input.Open(f).read()

    def WriteProfile(self, system_map, vtypes):
        # Sorting the json keys usually achieves much smaller file size due to
        # better compression. Its worth doing it once on conversion.
        self.output.StoreData("Constants.json", system_map)
        self.output.StoreData("vtypes.json", vtypes)
        self.output.StoreData("metadata", dict(
                ProfileClass=self.profile_class,
                Constants="Constants.json",
                VTypes="vtypes.json"))

    def Convert(self):
        raise RuntimeError("Unknown profile format.")

class LinuxConverter(ProfileConverter):
    """Convert an existing Linux profile zip file.

    Since building the linux profile often happens on the target system, where
    Rekall is not normall running, we just convert the result of running Make in
    the tools/linux/ directory. See tools/linux/README for details.

    In short:

    - Run make in tools/linux/ directory. This will build module_dwarf.ko with
      debugging symbols.

    - If you have zip installed, the above step will create the required zip
      file. Otherwise Create a zip file manually with module_dwarf.ko and
      /boot/System.map-`uname -r` (Sometimes when running make not as the root
      user, its not possible to read the System.map file).

    Finally use this tool to convert the profile to a Rekall compatible profile.
    """
    BASE_PROFILE_CLASS = "Linux"

    def ParseSystemMap(self, system_map):
        """Parse the system map and return a list of offset, symbol_name."""
        sys_map = {}
        # get the system map
        for line in system_map.splitlines():
            (address, _, symbol) = line.strip().split()
            try:
                sys_map[symbol] = long(address, 16)
            except ValueError:
                pass

        return sys_map

    def WriteProfile(self, system_map, vtypes):
        """Write all the components needed for the output profile."""
        # Try to guess the bit size of the system if not provided.
        if self.profile_class is None:
            largest_offset = max(system_map.values())
            if largest_offset > 2**32:
                self.profile_class = "%s64" % self.BASE_PROFILE_CLASS
            else:
                self.profile_class = "%s32" % self.BASE_PROFILE_CLASS

        super(LinuxConverter, self).WriteProfile(system_map, vtypes)

    def Convert(self):
        # Check for a linux profile. It should have a System.map in it.
        system_map = self.SelectFile("(^|/)System.map")
        if system_map:
            # Parse the system map file.
            system_map = self.ParseSystemMap(system_map)

            ko_file = self.SelectFile(r"\.ko$")
            if ko_file:
                logging.info("Converting Linux profile with ko module.")
                parser = dwarfparser.DWARFParser(StringIO.StringIO(ko_file))

                # Also write the ko file to ensure we get to keep it.
                with self.output.Create("module.ko") as fd:
                    fd.write(ko_file)

                return self.WriteProfile(system_map, parser.VType())

            dwarf_file = self.SelectFile(r"\.dwarf$")
            if dwarf_file:
                logging.info("Converting Linux profile with dwarf dump output")
                parser = dwarfdump.DWARFParser()
                for line in dwarf_file.splitlines():
                    parser.feed_line(line)

                # The dwarfdump module returns python code so we must exec it.
                l = {}
                exec(parser.print_output(), {}, l)

                return self.WriteProfile(system_map, l["linux_types"])

            # This is here just so we can transform Rekall profiles to Rekall
            # profiles.
            json_file = self.SelectFile(r"\.json$")
            if json_file:
                logging.info("Converting Linux profile with json vtype.")
                return self.WriteProfile(system_map, json.loads(json_file))

        raise RuntimeError("Unknown profile format.")


class OSXConverter(LinuxConverter):
    """Automatic converted for Volatility OSX style profiles.

    You can generate one of those using the instructions here:
    http://code.google.com/p/volatility/wiki/MacMemoryForensics#Building_a_Profile
    """
    BASE_PROFILE_CLASS = "Darwin"

    DLSYM_REGEX = re.compile("([^ ]+) '([^ ]+)'$")

    def ParseSystemMap(self, system_map):
        sys_map = {}
        for line in system_map.splitlines():
            if self.profile_class is None and "Symbol table for" in line:
                last_part = line.split()[-1]
                if last_part == "(x86_64)":
                    self.profile_class = "Darwin64"
                elif last_part == "(i386)":
                    self.profile_class = "Darwin32"
                else:
                    raise RuntimeError(
                        "Unknown Darwin Architecture %s" % last_part)

            # We only care about few things like functions and global symbols.
            if "N_FUN" in line or "EXT" in line or "N_STSYM" in line:
                m = self.DLSYM_REGEX.search(line)
                if m:
                    try:
                        sys_map[m.group(2)] = long(m.group(1), 16)
                    except ValueError:
                        pass

        return sys_map

    def Convert(self):
        # Check for an OSX profile.
        system_map = self.SelectFile("dsymutil$")
        if system_map:
            # Parse the system map file.
            system_map = self.ParseSystemMap(system_map)

            vtype_file = self.SelectFile(r"\.vtypes$")
            if vtype_file:
                logging.info(
                    "Converting Darwin profile with vtypes dump output")

                # The dwarfdump module returns python code so we must exec it.
                l = {}
                exec(vtype_file, {}, l)

                return self.WriteProfile(system_map, l["mac_types"])

            # This is here just so we can transform Rekall profiles to Rekall
            # profiles.
            json_file = self.SelectFile(r"\.json$")
            if json_file:
                logging.info("Converting Darwin profile with json vtype.")
                return self.WriteProfile(system_map, json.loads(json_file))

        raise RuntimeError("Unknown profile format.")


class WindowsConverter(ProfileConverter):
    """A converter from Volatility windows profiles.

    This converter must be manually specified.
    """

    def Convert(self):
        if not self.profile_class:
            raise RuntimeError("Profile class implementation not provided.")

        # The input file is a python file with a data structure in it.
        with open(self.input, "rb") as fd:
            l = {}
            exec(fd.read(), {}, l)

        return self.WriteProfile({}, l["ntkrnlmp_types"])


class ConvertProfile(plugin.Command):
    """Convert a profile from another program to the Rekall format.

    The Rekall profile format is optimized for loading at runtime. This plugin
    produces a Rekall profile from a variety of sources, including:

    - Linux debug compiled kernel module (see tool/linux/README)
    - OSX Dwarfdump outputs.
    """

    __name = "convert_profile"

    @classmethod
    def args(cls, parser):
        """Declare the command line args we need."""
        super(ConvertProfile, cls).args(parser)

        parser.add_argument(
            "--profile_class", default=None,
            help="The name of the profile implementation to specify. "
            "If not specified, we autodetect.")

        parser.add_argument(
            "--converter", default=None,
            help="The name of the converter to use. "
            "If not specified autoguess.")

        parser.add_argument("source",
                            help="Filename of profile to read.")

        parser.add_argument("destination",
                            help="Filename of profile to write.")

    def __init__(self, profile_class=None, converter=None, source=None,
                 destination=None, **kwargs):
        super(ConvertProfile, self).__init__(**kwargs)
        self.profile_class = profile_class
        self.converter = converter
        self.source = source
        self.destination = destination

    def ConvertProfile(self, input, output):
        """Converts the input profile to a new standard profile in output."""
        # First detect what kind of profile the input profile is.
        for converter in (LinuxConverter, OSXConverter):
            try:
                converter(input, output).Convert()
                logging.info("Converted %s to %s", input, output)
                return
            except RuntimeError:
                pass

        raise RuntimeError(
            "No suitable converter found - profile not recognized.")

    def render(self, renderer):
        try:
            output = io_manager.Factory(self.destination, mode="w")
        except IOError:
            logging.critical("Output profile File %s could not be opened.",
                             self.destination)
            return

        if self.converter:
            cls = ProfileConverter.classes.get(self.converter)
            if not cls:
                raise IOError("Unknown converter %s" % self.converter)

            return cls(self.source, output,
                       profile_class=self.profile_class).Convert()

        try:
            input = io_manager.Factory(self.source, mode="r")
        except IOError:
            logging.critical("Input profile file %s could not be opened.",
                             self.source)
            return

        with input, output:
            self.ConvertProfile(input, output)




class ProfileCompressor(object):
    def __init__(self):
        self.data = {}

class CompressProfile(plugin.Command):
    """Reprocess a profile db to produce a new compressed profile container.

    The Rekall builtin profile container contains only the information directly
    used by various plugins. In order to record all the vtype access that
    plugins make, set the environment variable DEBUG_PROFILE to point at a path
    where a json database will be maintained.

    Periodically running this plugin will rebuild the compressed builtin profile
    container by adding further information from this database.
    """
    __name = "update_builtin_profile"

    @classmethod
    def is_active(cls, session):
        """Only active when the DEBUG_PROFILE environment variable is set."""
        return bool(os.environ.get("DEBUG_PROFILE"))

    @classmethod
    def args(cls, parser):
        super(CompressProfile, cls).args(parser)
        parser.add_argument(
            "--output", default="rekall/builtin_profiles.py",
            help="The path to the builtin profile file to be produced.")

    def __init__(self, output=None, **kwargs):
        super(CompressProfile, self).__init__(**kwargs)
        self.output = output
        self.database = os.environ.get("DEBUG_PROFILE")
        self.data = builtin_profiles.BuiltInProfiles.data.copy()

    def Compress(self, compression_db, profile):
        dest = self.data[profile.name] = dict(
            metadata={
            'Constants': 'Constants.json',
            'ProfileClass': profile.__class__.__name__,
            'VTypes': 'vtypes.json',
            })

        constants = compression_db.get("Constants")
        if constants:
            self.CompressConstants(profile.constants, constants, dest)

        self.CompressVTypes(profile.vtypes, compression_db, dest)

    def CompressVTypes(self, profile_vtypes, compression_db, dest):
        for struct, members in compression_db.items():
            profile_struct = profile_vtypes.get(struct)
            if profile_struct:
                size, profile_desc = profile_struct
                out = {}
                dest.setdefault("vtypes.json", {})[struct] = [size, out]

                for member in members:
                    if member in profile_desc:
                        out[member] = profile_desc[member]

    def CompressConstants(self, profile_constants, constants, dest):
        out = dest["Constants.json"] = {}
        for constant in constants:
            if constant in profile_constants:
                out[constant] = profile_constants[constant]

    def PPrint(self, data, depth=0):
        """A pretty printer for a profile.

        This only supports dict, list and non-unicode strings.
        """
        result = []
        if isinstance(data, dict):
            result.append("{")
            for key in sorted(data):
                key = str(key)

                result.append(
                    " %r: %s," % (
                        key, self.PPrint(data[key], depth + 1).strip()))

            result.append("}")

            return "\n".join([(" " * depth + x) for x in result])

        if isinstance(data, list):
            for item in data:
                result.append(self.PPrint(item, depth).strip())

            return "[" + ", ".join(result) + "]"

        if isinstance(data, basestring):
            try:
                data = data.encode("ascii")
            except UnicodeError:
                pass

        return repr(data)

    CODE_TEMPLATE = """
# AUTOGENERATED File do not edit!
# File generated by the profile_tool.py 'compress' command.

from rekall import io_manager


class BuiltInProfiles(io_manager.BuiltInManager):
   data = %s

"""

    def render(self, renderer):
        try:
            # Get the profile access database.
            db = json.load(open(self.database, "rb"))
        except IOError as e:
            logging.error("Unable to open database file: %s", e)
            return

        for name, desc in db.items():
            # Try to open the profile named in the database.
            try:
                source_profile = self.session.LoadProfile(name)
            except ValueError:
                continue

            if source_profile:
                self.Compress(desc, source_profile)

        with open(self.output, "wb") as fd:
            fd.write(self.CODE_TEMPLATE % self.PPrint(self.data))
            logging.info("Wrote compressed profile into %s", self.destination)



#!/usr/bin/python

# Rekall Memory Forensics
# Copyright (C) 2012 Michael Cohen <scudette@gmail.com>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
"""This module manages the command line parsing logic."""

__author__ = "Michael Cohen <scudette@gmail.com>"

import argparse
import logging
import re
import os
import sys
import zipfile

from rekall import config
from rekall import constants
from rekall import plugin


config.DeclareOption("--plugin", default=[], nargs="+",
                     help="Load user provided plugin bundle.")

config.DeclareOption("--output",
                     help="Write to this output file.")

config.DeclareOption(
    "-h", "--help", default=False, action="store_true",
    help="Show help about global paramters.")

config.DeclareOption("--overwrite", action="store_true", default=False,
                     help="Allow overwriting of output files.")


class IntParser(argparse.Action):
    """Class to parse ints either in hex or as ints."""
    def parse_int(self, value):
        # Support suffixes
        multiplier = 1
        m = re.search("(.*)(mb|kb|m|k)", value)
        if m:
            value = m.group(1)
            suffix = m.group(2).lower()
            if suffix in ("mb", "m"):
                multiplier = 1024 * 1024
            elif suffix in ("kb", "k"):
                multiplier = 1024

        try:
            if value.startswith("0x"):
                value = int(value, 16) * multiplier
            else:
                value = int(value) * multiplier
        except ValueError:
            raise argparse.ArgumentError(self, "Invalid integer value")

        return value

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.parse_int(values))


class ArrayIntParser(IntParser):
    """Parse input as a comma separated list of integers.

    We support input in the following forms:

    --pid 1,2,3,4,5

    --pid 1 2 3 4 5

    --pid 0x1 0x2 0x3
    """

    def __call__(self, parser, namespace, values, option_string=None):
        result = []
        if isinstance(values, basestring):
            values = [values]

        for value in values:
            result.extend([self.parse_int(x) for x in value.split(",")])

        setattr(namespace, self.dest, result)


class ArrayStringParser(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        result = []
        if isinstance(values, basestring):
            values = [values]

        for value in values:
            result.extend([x for x in value.split(",")])

        setattr(namespace, self.dest, result)


class MockArgParser(object):
    def add_argument(self, short_flag="", long_flag="", dest="", **_):
        if short_flag.startswith("--"):
            flag = short_flag
        elif long_flag.startswith("--"):
            flag = long_flag
        elif dest:
            flag = dest
        else:
            flag = short_flag

        # This function will be called by the args() class method, and we just
        # keep track of the args this module defines.
        arg_name = flag.strip("-").replace("-", "_")

        self.args[arg_name] = None

    def build_args_dict(self, cls, namespace):
        """Build a dict suitable for **kwargs from the namespace."""
        self.args = {}

        # Discover all the args this module uses.
        cls.args(self)

        for key in self.args:
            self.args[key] = getattr(namespace, key)

        return self.args


class RekallArgParser(argparse.ArgumentParser):
    ignore_errors = False

    def error(self, message):
        if self.ignore_errors:
            return

        # We trap this error especially since we launch the volshell.
        if message == "too few arguments":
            return

        super(RekallArgParser, self).error(message)

    def parse_known_args(self, args=None, namespace=None, force=False, **_):
        self.ignore_errors = force

        result = super(RekallArgParser, self).parse_known_args(
            args=args, namespace=namespace)

        return result

    def print_help(self, file=None):
        if self.ignore_errors:
            return

        return super(RekallArgParser, self).print_help(file=file)

    def exit(self, *args, **kwargs):
        if self.ignore_errors:
            return

        return super(RekallArgParser, self).exit(*args, **kwargs)


def LoadPlugins(paths=None):
    PYTHON_EXTENSIONS = [".py", ".pyo", ".pyc"]

    for path in paths:
        if not os.access(path, os.R_OK):
            logging.error("Unable to find %s", path)
            continue

        path = os.path.abspath(path)
        directory, filename = os.path.split(path)
        module_name, ext = os.path.splitext(filename)

        # Its a python file.
        if ext in PYTHON_EXTENSIONS:
            # Make sure python can find the file.
            sys.path.insert(0, directory)

            try:
                logging.info("Loading user plugin %s", path)
                __import__(module_name)
            except Exception, e:
                logging.error("Error loading user plugin %s: %s", path, e)
            finally:
                sys.path.pop(0)

        elif ext == ".zip":
            zfile = zipfile.ZipFile(path)

            # Make sure python can find the file.
            sys.path.insert(0, path)
            try:
                logging.info("Loading user plugin archive %s", path)
                for name in zfile.namelist():
                    # Change from filename to python package name.
                    module_name, ext = os.path.splitext(name)
                    if ext in PYTHON_EXTENSIONS:
                        module_name = module_name.replace("/", ".").replace(
                            "\\", ".")

                        try:
                            __import__(module_name.strip("\\/"))
                        except Exception as e:
                            logging.error("Error loading user plugin %s: %s",
                                          path, e)

            finally:
                sys.path.pop(0)

        else:
            logging.error("Plugin %s has incorrect extension.", path)


def LoadProfileIntoSession(parser, argv, user_session):
    # Figure out the profile
    known_args, _ = parser.parse_known_args(args=argv)

    # Force debug level logging with the verbose flag.
    if getattr(known_args, "verbose", None):
        known_args.logging = "DEBUG"

    with user_session.state as state:
        config.MergeConfigOptions(state)

        for arg, value in known_args.__dict__.items():
            state.Set(arg, value)

    if user_session.profile is None:
        guesser = user_session.plugins.guess_profile()
        guesser.update_session()

    # Now load the third party user plugins. These may introduce additional
    # plugins with args.
    LoadPlugins(user_session.state.plugin)


def parse_args(argv=None, user_session=None):
    """Parse the args from the command line argv."""
    parser = RekallArgParser(
        description=constants.BANNER,
        conflict_handler='resolve',
        add_help=False,
        epilog='When no module is provided, drops into interactive mode',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    config.RegisterArgParser(parser)

    # First load the profile to enable the module selection (which depends on
    # the profile).
    LoadProfileIntoSession(parser, argv, user_session)

    # Add module specific args.
    subparsers = parser.add_subparsers(
        description="The following plugins can be selected.",
        metavar='Plugin',
        )

    parsers = {}

    # Add module specific parser for each module.
    classes = []
    for cls in plugin.Command.classes.values():
        if (cls.name and cls.is_active(user_session) and not
            cls.interactive):
            classes.append(cls)

    for cls in sorted(classes, key=lambda x: x.name):
        docstring = cls.__doc__ or " No Docs "
        doc = docstring.splitlines()[0] or " No Docs "
        name = cls.name
        try:
            module_parser = parsers[name]
        except KeyError:
            parsers[name] = module_parser = subparsers.add_parser(
                cls.name, help=doc, description=docstring)

            cls.args(module_parser)
            module_parser.set_defaults(module=cls.name)

    # Parse the final command line.
    result = parser.parse_args(argv)

    # We handle help especially since we want to enumerate all plugins.
    if getattr(result, "help", None):
        parser.print_help()
        sys.exit(-1)

    return result




# Rekall Memory Forensics
# Copyright (c) 2008-2011 Volatile Systems
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
@author:       Bradley L Schatz
@license:      GNU General Public License 2.0 or later
@contact:      bradley@schatzforensic.com.au

This file provides support for windows Windows 7 SP 0.
"""

#pylint: disable-msg=C0111

import windows
from rekall import obj
from rekall.plugins.overlays import basic
from rekall.plugins.overlays.windows import windows


# In windows 7 the VadRoot is actually composed from _MMADDRESS_NODEs instead of
# _MMVAD structs.
win7_overlays = {
    '_EPROCESS': [ None, {
            # A symbolic link to the real vad root.
            'RealVadRoot': lambda x: x.VadRoot.BalancedRoot
            }],

    '_MMADDRESS_NODE': [ None, {
            'Tag': [-12, ['String', dict(length=4)]],
            }],

    '_MMVAD_SHORT': [ None, {
            'Tag': [-12 , ['String', dict(length = 4)]],
            'Start': lambda x: x.StartingVpn << 12,
            'End': lambda x: ((x.EndingVpn + 1) << 12) - 1,
            'Length': lambda x: x.End - x.Start + 1,
            'CommitCharge': lambda x: x.u.VadFlags.CommitCharge,
            }],

    '_MMVAD': [ None, {
            'Tag': [-12 , ['String', dict(length = 4)]],
            'ControlArea': lambda x: x.Subsection.ControlArea,
            'Start': lambda x: x.StartingVpn << 12,
            'End': lambda x: ((x.EndingVpn + 1) << 12) - 1,
            'Length': lambda x: x.End - x.Start + 1,
            'CommitCharge': lambda x: x.u.VadFlags.CommitCharge,
            }],

    '_MMVAD_LONG': [ None, {
            'Tag': [-12 , ['String', dict(length = 4)]],
            'ControlArea': lambda x: x.Subsection.ControlArea,
            'Start': lambda x: x.StartingVpn << 12,
            'End': lambda x: ((x.EndingVpn + 1) << 12) - 1,
            'Length': lambda x: x.End - x.Start + 1,
            'CommitCharge': lambda x: x.u.VadFlags.CommitCharge,
            }],

    "_CONTROL_AREA": [None, {
            'FilePointer': [None, ['_EX_FAST_REF', dict(target="_FILE_OBJECT")]],
            }],
    }


class _OBJECT_HEADER(windows._OBJECT_HEADER):
    """A Rekall Memory Forensics object to handle Windows 7 object headers.

    Windows 7 changes the way objects are handled:
    References: http://www.codemachine.com/article_objectheader.html

    The following debugger command find the type object for index 5:
    dt nt!_OBJECT_TYPE poi(nt!ObTypeIndexTable + ( 5 * @$ptrsize ))
    """
    type_map = { 2: 'Type',
                3: 'Directory',
                4: 'SymbolicLink',
                5: 'Token',
                6: 'Job',
                7: 'Process',
                8: 'Thread',
                9: 'UserApcReserve',
                10: 'IoCompletionReserve',
                11: 'DebugObject',
                12: 'Event',
                13: 'EventPair',
                14: 'Mutant',
                15: 'Callback',
                16: 'Semaphore',
                17: 'Timer',
                18: 'Profile',
                19: 'KeyedEvent',
                20: 'WindowStation',
                21: 'Desktop',
                22: 'TpWorkerFactory',
                23: 'Adapter',
                24: 'Controller',
                25: 'Device',
                26: 'Driver',
                27: 'IoCompletion',
                28: 'File',
                29: 'TmTm',
                30: 'TmTx',
                31: 'TmRm',
                32: 'TmEn',
                33: 'Section',
                34: 'Session',
                35: 'Key',
                36: 'ALPC Port',
                37: 'PowerRequest',
                38: 'WmiGuid',
                39: 'EtwRegistration',
                40: 'EtwConsumer',
                41: 'FilterConnectionPort',
                42: 'FilterCommunicationPort',
                43: 'PcwObject',
            }

    # This specifies the order the headers are found below the _OBJECT_HEADER
    optional_header_mask = (('CreatorInfo', '_OBJECT_HEADER_CREATOR_INFO', 0x01),
                            ('NameInfo', '_OBJECT_HEADER_NAME_INFO', 0x02),
                            ('HandleInfo', '_OBJECT_HEADER_HANDLE_INFO', 0x04),
                            ('QuotaInfo', '_OBJECT_HEADER_QUOTA_INFO', 0x08),
                            ('ProcessInfo', '_OBJECT_HEADER_PROCESS_INFO', 0x10))

    def find_optional_headers(self):
        """Find this object's optional headers."""
        offset = self.obj_offset
        info_mask = int(self.InfoMask)

        for name, struct, mask in self.optional_header_mask:
            if info_mask & mask:
                offset -= self.obj_profile.get_obj_size(struct)
                o = self.obj_profile.Object(type_name=struct, offset=offset, vm=self.obj_vm)
                self._preamble_size += o.size()
            else:
                o = obj.NoneObject("Header not set")

            setattr(self, name, o)

    def get_object_type(self, vm=None):
        """Return the object's type as a string"""
        return self.type_map.get(self.TypeIndex.v(), '')

    def is_valid(self):
        """Determine if the object makes sense."""
        # These need to be reasonable.
        if (self.PointerCount < 0x100000 and self.HandleCount < 0x1000 and
            self.PointerCount >= 0 and self.HandleCount >= 0 and
            self.TypeIndex <= len(self.type_map) and
            self.TypeIndex > 0):
            return True

        return False


class _MMADDRESS_NODE(windows.VadTraverser):
    """In win7 the base of all Vad objects in _MMADDRESS_NODE.

    The Vad structures can be either _MMVAD_SHORT or _MMVAD or _MMVAD_LONG. At
    the base of each struct there is an _MMADDRESS_NODE which contains the
    LeftChild and RightChild members. In order to traverse the tree, we follow
    the _MMADDRESS_NODE and create the required _MMVAD type at each point
    depending on their tags.
    """

    ## The actual type depends on this tag value.
    tag_map = {'Vadl': '_MMVAD_LONG',
               'VadS': '_MMVAD_SHORT',
               'Vad ': '_MMVAD',
               'VadF': '_MMVAD_SHORT',
               'Vadm': '_MMVAD_LONG',
              }


class _POOL_HEADER(windows._POOL_HEADER):
    """A class for pool headers"""

    @property
    def NonPagedPool(self):
        return self.PoolType.v() % 2 == 0 and self.PoolType.v() > 0

    @property
    def PagedPool(self):
        return self.PoolType.v() % 2 == 1


class Win7BaseProfile(windows.BaseWindowsProfile):
    """The common ancestor of all windows 7 profiles."""

    __abstract = True

    def __init__(self, **kwargs):
        super(Win7BaseProfile, self).__init__(**kwargs)
        self.add_types({
                'pointer64': ['NativeType', dict(format_string='<Q')]
                })
        self.add_overlay(win7_overlays)

        self.add_classes(dict(_OBJECT_HEADER=_OBJECT_HEADER,
                              _MMADDRESS_NODE=_MMADDRESS_NODE,
                              _POOL_HEADER=_POOL_HEADER,
                              pointer64=obj.Pointer))


class Win7SP0x86(basic.Profile32Bits, Win7BaseProfile):
    """ A Profile for Windows 7 SP0 x86 """
    _md_major = 6
    _md_minor = 1
    _md_build = 7600
    _md_type = "Kernel"

    def __init__(self, **kwargs):
        super(Win7SP0x86, self).__init__(**kwargs)

        # Import the actual vtypes on demand here to reduce memory usage.
        from rekall.plugins.overlays.windows import win7_sp0_x86_vtypes

        self.add_types(win7_sp0_x86_vtypes.ntkrnlmp_types)


class Win7SP0x64(basic.ProfileLLP64, Win7BaseProfile):
    """ A Profile for Windows 7 SP0 x64 """
    _md_major = 6
    _md_minor = 1
    _md_build = 7600
    _md_type = "Kernel"


class Win7SP1x86(basic.Profile32Bits, Win7BaseProfile):
    """ A Profile for Windows 7 SP1 x86 """
    _md_major = 6
    _md_minor = 1
    _md_build = 7601
    _md_type = "Kernel"


class Win7SP1x64(basic.ProfileLLP64, Win7BaseProfile):
    """ A Profile for Windows 7 SP1 x64 """
    _md_major = 6
    _md_minor = 1
    _md_build = 7601
    _md_type = "Kernel"



from rekall.plugins.linux import arp
from rekall.plugins.linux import bash
from rekall.plugins.linux import check_afinfo
from rekall.plugins.linux import check_creds
from rekall.plugins.linux import check_idt
from rekall.plugins.linux import check_fops
from rekall.plugins.linux import check_modules
from rekall.plugins.linux import check_syscall
from rekall.plugins.linux import check_tty
from rekall.plugins.linux import common
from rekall.plugins.linux import cpuinfo
from rekall.plugins.linux import dmesg
from rekall.plugins.linux import ifconfig
from rekall.plugins.linux import iomem
from rekall.plugins.linux import lsmod
#from rekall.plugins.linux import mount
from rekall.plugins.linux import netstat
from rekall.plugins.linux import lsof
from rekall.plugins.linux import pas2kas
from rekall.plugins.linux import proc_maps
from rekall.plugins.linux import psaux
from rekall.plugins.linux import pslist
from rekall.plugins.linux import pstree



# Rekall Memory Forensics
# Copyright (C) 2010 Brendan Dolan-Gavitt
# Copyright (c) 2011 Michael Cohen <scudette@gmail.com>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
@author:       Brendan Dolan-Gavitt
@license:      GNU General Public License 2.0 or later
@contact:      brendandg@gatech.edu
@organization: Georgia Institute of Technology
"""
import logging
import json
import posixpath
import re
import sys
import zipfile
import StringIO

from rekall import obj
from rekall import plugin
from rekall import utils

from rekall.plugins.overlays import basic
from rekall.plugins.overlays.linux import vfs
from rekall.plugins.overlays.linux import dwarfdump

# Try to use the elftools directly if they are present.
try:
    from rekall.plugins.overlays.linux import dwarfparser

    logging.info("Unable to load the dwarfparser module. Do you have "
                 "elftools installed?")
except ImportError:
    dwarfparser = None

linux_overlay = {
    'task_struct' : [None, {
            'comm': [None , ['UnicodeString', dict(length = 16)]],
            'uid': lambda x: x.m("uid") or x.cred.uid,
            'gid': lambda x: x.m("gid") or x.cred.gid,
            'euid': lambda x: x.m("euid") or x.cred.euid,
            }],

    'module' : [None, {
            'name': [None , ['UnicodeString', dict(length = 60)]],
            'kp': [None, ['Pointer', dict(
                        target='Array',
                        target_args=dict(
                            target='kernel_param',
                            count=lambda x: x.num_kp))]],
            }],

    'kernel_param': [None, {
            'name' : [None , ['Pointer', dict(target='UnicodeString')]],
            'getter_addr': lambda x: (x.m("get") or x.ops.get),
            }],

    'kparam_array': [None, {
            'getter_addr': lambda x: (x.m("get") or x.ops.get),
            }],

    'super_block' : [None, {
        's_id' : [None , ['UnicodeString', dict(length = 32)]],
        'major': lambda x: x.s_dev >> 20,
        'minor': lambda x: x.s_dev & ((1 << 20) - 1),
        }],

    'net_device'  : [None, {
            # Flags defined in include/linux/if.h
            'flags': [None, ['Flags', dict(
                        maskmap={
                            "IFF_UP":           0x1,  # interface is up
                            "IFF_BROADCAST":    0x2,  # broadcast address valid
                            "IFF_DEBUG":        0x4,  # turn on debugging
                            "IFF_LOOPBACK":     0x8,  # is a loopback net
                            "IFF_POINTOPOINT": 0x10,  # interface is has p-p link
                            "IFF_NOTRAILERS":  0x20,  # avoid use of trailers
                            "IFF_RUNNING":     0x40,  # interface RFC2863 OPER_UP
                            "IFF_NOARP":       0x80,  # no ARP protocol
                            "IFF_PROMISC":     0x100, # receive all packets
                            "IFF_ALLMULTI":    0x200,
                            }
                        )]],

            'name' : [None , ['UnicodeString', dict(length = 16)]],

            'mac_addr': lambda x: (x.perm_addr or x.dev_addr).cast("MacAddress"),

            'ip_ptr': [None, ['Pointer', dict(target="in_device")]],
            }],

    'in_ifaddr': [None, {
            'ifa_address': [None, ['IpAddress']],
            'ifa_label': [None, ['String']],
            }],

    'sockaddr_un' : [None, {
        'sun_path'      : [ None , ['UnicodeString', dict(length =108)]],
        }],

    'cpuinfo_x86' : [None, {
        'x86_model_id'  : [ None , ['UnicodeString', dict(length = 64)]],
        'x86_vendor_id' : [ None,  ['UnicodeString', dict(length = 16)]],
        }],

    'module_sect_attr': [None, {
            'name': [None, ['Pointer', dict(target='UnicodeString')]]
            }],

    # The size of the log record is stored in the len member.
    'log': [lambda x: x.len, {
            # The log message starts after the level member.
            'message': [lambda x: x.flags.obj_offset + x.flags.size(),
                        ['UnicodeString', dict(

                        # The length of the message is the text and an optional
                        # dict.
                        length=lambda x: x.text_len + x.dict_len)]],

            'level': [None, ['Enumeration', {
                        'choices': {
                            0: 'LOG_EMERG',
                            1: 'LOG_ALERT',
                            2: 'LOG_CRIT',
                            3: 'LOG_ERR',
                            4: 'LOG_WARNING',
                            5: 'LOG_NOTICE',
                            6: 'LOG_INFO',
                            7: 'LOG_DEBUG'
                            },
                        'target': 'BitField',
                        'target_args': dict(
                            start_bit=0, end_bit=3),
                        }]],
            }],

    "file": [None, {
            "dentry": lambda x: x.m("f_dentry") or x.f_path.dentry,
            "vfsmnt": lambda x: x.m("f_vfsmnt") or x.f_path.mnt,
            }],

    'vm_area_struct' : [ None, {
            'vm_flags' : [None, ['PermissionFlags', dict(
                        bitmap={
                            'r': 0,
                            'w': 1,
                            'x': 2
                            }
                        )]],
            }],

    'dentry': [None, {
            'd_flags': [None, ['Flags', dict(
                        maskmap={
                            # /* autofs: "under construction" */
                            "DCACHE_AUTOFS_PENDING": 0x0001,
                            #     /* this dentry has been "silly renamed" and
                            #     has to be deleted on the last dput() */
                            "DCACHE_NFSFS_RENAMED":  0x0002,
                            "DCACHE_DISCONNECTED":  0x0004,
                            #  /* Recently used, don't discard. */
                            "DCACHE_REFERENCED": 0x0008,
                            "DCACHE_UNHASHED": 0x0010,
                            #  /* Parent inode is watched by inotify */
                            "DCACHE_INOTIFY_PARENT_WATCHED": 0x0020,
                            #  /* Parent inode is watched by some fsnotify
                            #  listener */
                            "DCACHE_FSNOTIFY_PARENT_WATCHED":  0x0080,
                            },
                        target="unsigned int",
                        )]],
            'is_root': lambda x: x == x.d_parent,
            }],

    'qstr': [None, {
            'name': [None, ['Pointer', dict(
                        target='UnicodeString',
                        target_args=dict(
                            length=lambda x: x.m("len") or x.u1.u1.len,
                            )
                        )]],
            }],

    "files_struct": [None, {
            # Version independent pointer to the fd table.
            '_fd': lambda x: x.m("fdt").fd or x.fd,

            # The fds table is an array of pointers to file structs. The index
            # into the array is known in user space as a file descriptor. If the
            # pointer in the array is invalid, the file descriptor is closed.
            'fds': lambda x: x._fd.dereference_as(
                "Array",
                target_args=dict(
                    target="Pointer",
                    target_args=dict(
                        target="file"
                        ),
                    count=lambda x: x.m("fdt").max_fds or x.max_fds,
                    )
                )
            }],
    "proc_dir_entry": [None, {
            'Name': lambda x: (
                # 2.6 kernel.
                x.name.cast("Pointer", target="String").deref() or

                # 3.x kernel.
                x.name.cast("String")),
            }],

    "kobject": [None, {
            'name': [None, ["Pointer", dict(
                        target="String",
                        target_args=dict(length=32),
                        )]],
            }],

    'gate_struct64': [None, {
            'Address': lambda x: (x.offset_low |
                                  x.offset_middle << 16 |
                                  x.offset_high << 32),
            }],

    'desc_struct': [None, {
            'Address': lambda x: (x.b & 0xffff0000) | (x.a & 0x0000ffff),
            }],

    'tty_driver': [None, {

            "name": [None, ["Pointer", dict(
                        target="String"
                        )]],

            "ttys": [None, ["Pointer", dict(
                        target="Array",
                        target_args=dict(
                            count=lambda x: x.num,
                            target="Pointer",
                            target_args=dict(
                                target="tty_struct"
                                )
                            )
                        )]],
            }],

    'tty_struct': [None, {
            'name': [None, ["String"]],
            }],

    "resource": [None, {
            "name": [None, ["Pointer", dict(
                        target="String"
                        )]],
            }],
    "sock": [None, {
            # http://lxr.free-electrons.com/source/include/linux/net.h#L58
            "sk_type": [None, ["Enumeration", dict(
                        choices={1:"SOCK_STREAM",
                                 2:"SOCK_DGRAM",
                                 3:"SOCK_RAW",
                                 4:"SOCK_RDM",
                                 5:"SOCK_SEQPACKET",
                                 6:"SOCK_DCCP",
                                 10:"SOCK_PACKET",
                                 },
                        target="BitField",
                        target_args=dict(
                            start_bit=16, end_bit=32))]],

            "sk_protocol": [None, ["Enumeration", dict(
                        # http://lxr.free-electrons.com/source/include/uapi/linux/in.h?#L26
                        choices={
                            0:"IPPROTO_HOPOPT",               # Dummy protocol forTCP
                            1:"IPPROTO_ICMP",             # Internet Control Message Protocol
                            2:"IPPROTO_IGMP",             # Internet Group Management Protocol
                            4:"IPPROTO_IPV4",             # IPIP tunnels (older KA9Q tunnels use 94)
                            6:"IPPROTO_TCP",              # Transmission Control Protocol
                            8:"IPPROTO_EGP",              # Exterior Gateway Protocol
                            12:"IPPROTO_PUP",             # PUP protocol
                            17:"IPPROTO_UDP",             # User Datagram Protocol
                            22:"IPPROTO_IDP",             # XNS IDP protocol
                            29:"IPPROTO_TP",              # SO Transport Protocol Class 4
                            33:"IPPROTO_DCCP",            # Datagram Congestion Control Protocol
                            41:"IPPROTO_IPV6",            # IPv6-in-IPv4 tunnelling
                            46:"IPPROTO_RSVP",            # RSVP Protocol
                            47:"IPPROTO_GRE",             # Cisco GRE tunnels (rfc 1701",1702)
                            50:"IPPROTO_ESP",             # Encapsulation Security Payload protocol
                            51:"IPPROTO_AH",              # Authentication Header protocol
                            92:"IPPROTO_MTP",             # Multicast Transport Protocol
                            94:"IPPROTO_BEETPH",          # IP option pseudo header for BEET
                            98:"IPPROTO_ENCAP",           # Encapsulation Header
                            103:"IPPROTO_PIM",            # Protocol Independent Multicast
                            108:"IPPROTO_COMP",           # Compression Header Protocol
                            132:"IPPROTO_SCTP",           # Stream Control Transport Protocol
                            136:"IPPROTO_UDPLITE",        # UDP-Lite (RFC 3828)
                            255:"IPPROTO_RAW",            # Raw IP packets
                            },
                        target="BitField",
                        target_args=dict(
                            start_bit=8, end_bit=16),
                        )]],
            }],
    "sock_common": [None, {
            "skc_state": [None, ["Enumeration", dict(
                        # http://lxr.free-electrons.com/source/include/net/tcp_states.h#L16
                        choices={
                            1:"TCP_ESTABLISHED",
                            2:"TCP_SYN_SENT",
                            3:"TCP_SYN_RECV",
                            4:"TCP_FIN_WAIT1",
                            5:"TCP_FIN_WAIT2",
                            6:"TCP_TIME_WAIT",
                            7:"TCP_CLOSE",
                            8:"TCP_CLOSE_WAIT",
                            9:"TCP_LAST_ACK",
                            10:"TCP_LISTEN",
                            11:"TCP_CLOSING",
                            },
                        target="unsigned char",
                        )]],

            "skc_family": [None, ["Enumeration", dict(
                        # Partial List
                        # from: http://lxr.free-electrons.com/source/include/linux/socket.h#L140
                        choices={
                            0:"AF_UNSPEC",
                            1:"AF_UNIX",        # Unix domain sockets
                            2:"AF_INET",        # Internet IP Protocol
                            3:"AF_AX25",        # Amateur Radio AX.25
                            4:"AF_IPX",         # Novell IPX
                            5:"AF_APPLETALK",   # AppleTalk DDP
                            7:"AF_BRIDGE",      # Multiprotocol bridge
                            10:"AF_INET6",      # IP version 6
                            },
                        target="short unsigned int"
                        )]],
            }],
    }


class list_head(basic.ListMixIn, obj.Struct):
    """A list_head makes a doubly linked list."""
    _forward = "next"
    _backward = "prev"


class hlist_head(obj.Struct):
    def list_of_type(self, type, member):
        hlist = self.first.deref()
        while hlist:
            yield basic.container_of(hlist, type, member)

            hlist = hlist.next


class inet_sock(obj.Struct):
    """Class for an internet socket object"""

    @property
    def src_port(self):
        return (self.m("sport") or self.m("inet_sport")).cast(
            "unsigned be short")

    @property
    def dst_port(self):
        return (self.m("dport") or self.m("inet_dport")).cast(
            "unsigned be short") or self.sk.m("__sk_common").u3.u1.skc_dport

    @property
    def src_addr(self):
        if self.sk.m("__sk_common").skc_family == "AF_INET":
            return (self.m("rcv_saddr") or self.m("inet_rcv_saddr") or
                    self.sk.m("__sk_common").u1.u1.skc_rcv_saddr).cast(
                "IpAddress")

        else:
            return self.pinet6.saddr.cast("Ipv6Address")

    @property
    def dst_addr(self):
        if self.sk.m("__sk_common").skc_family == "AF_INET":
            return (self.m("daddr") or self.m("inet_daddr") or
                    self.sk.m("__sk_common").u1.u1.skc_daddr).cast(
                "IpAddress")

        else:
            return self.pinet6.daddr.cast("Ipv6Address")


class files_struct(obj.Struct):

    def get_fds(self):
        if hasattr(self, "fdt"):
            fdt = self.fdt
            ret = fdt.fd.dereference()
        else:
            ret = self.fd.dereference()

        return ret

    def get_max_fds(self):
        if hasattr(self, "fdt"):
            ret = self.fdt.max_fds
        else:
            ret = self.max_fds

        return ret


class dentry(obj.Struct):
    @property
    def path(self):
        dentry = self

        path_components = []

        # Check for deleted dentry.
        if self.d_flags.DCACHE_UNHASHED and not self.is_root:
            return " (deleted) "

        while len(path_components) < 50:
            if dentry.is_root:
                break

            component = utils.SmartUnicode(dentry.d_name.name.deref())
            path_components = [component] + path_components
            dentry = dentry.d_parent

        result = '/'.join(filter(None, path_components))

        if result.startswith(("socket:", "pipe:")):
            if result.find("]") == -1:
                result += ":[{0}]".format(self.d_inode.i_ino)

        elif result != "inotify":
            result = '/' + result

        return result


class task_struct(obj.Struct):

    @property
    def commandline(self):
        if self.mm:
            # The argv string is initialized inside the process's address space.
            proc_as = self.get_process_address_space()

            # read argv from userland
            argv = proc_as.read(self.mm.arg_start,
                                self.mm.arg_end - self.mm.arg_start)

            if argv:
                # split the \x00 buffer into args
                name = " ".join(argv.split("\x00"))
            else:
                name = ""
        else:
            # kernel thread
            name = "[" + self.comm + "]"

        return name

    def get_path(self, filp):
        """Resolve the dentry, vfsmount relative to this task's chroot.

        Returns:
          An absolute path to the global filesystem mount. (I.e. we do not
          truncate the path at the chroot point as the kernel does).
        """
        # The specific implementation depends on the kernel version.
        try:
            # Newer kernels have mnt_parent in the mount struct, not in the
            # vfsmount struct.
            self.obj_profile.get_obj_offset("vfsmount", "mnt_parent")

            return vfs.Linux26VFS().get_path(self, filp)
        except KeyError:
            return vfs.Linux3VFS().get_path(self, filp)

    def get_process_address_space(self):
        directory_table_base = self.obj_vm.vtop(self.mm.pgd.v())

        try:
            process_as = self.obj_vm.__class__(
                base=self.obj_vm.base, session=self.obj_vm.session,
                dtb = directory_table_base)

        except AssertionError, _e:
            return obj.NoneObject("Unable to get process AS")

        process_as.name = "Process {0}".format(self.pid)

        return process_as


class timespec(obj.Struct):
    # The following calculate the number of ns each tick is.
    # http://lxr.free-electrons.com/source/include/linux/jiffies.h?v=2.6.32#L12

    # The HZ value should be obtained from the auxilary vector but for now we
    # hard code it. TODO: http://lwn.net/Articles/519085/
    HZ = 1000

    # The clock frequency of the i8253/i8254 PIT
    CLOCK_TICK_RATE = PIT_TICK_RATE = 1193182

    # LATCH is used in the interval timer and ftape setup.
    LATCH = ((CLOCK_TICK_RATE + HZ/2) / HZ)

    # HZ is the requested value. ACTHZ is actual HZ
    ACTHZ = (CLOCK_TICK_RATE / LATCH)

    # TICK_NSEC is the time between ticks in nsec assuming real ACTHZ
    TICK_NSEC = 1000000 * 1000 /  ACTHZ

    NSEC_PER_SEC = 1000000000

    @property
    def wall_to_monotonic(self):
        wall_addr = self.obj_profile.get_constant("wall_to_monotonic")
        if wall_addr:
            return self.obj_profile.timespec(vm=self.obj_vm, offset=wall_addr)

        # After Kernel 3.3 wall_to_monotonic is stored inside the timekeeper.
        timekeeper_addr = self.obj_profile.get_constant("timekeeper")
        if timekeeper_addr:
            return  self.obj_profile.timekeeper(
                vm=self.obj_vm, offset=timekeeper_addr).wall_to_monotonic

    @property
    def total_sleep_time(self):
        total_sleep_time_addr = self.obj_profile.get_constant("total_sleep_time")
        if total_sleep_time_addr:
            return self.obj_profile.timespec(
                vm=self.obj_vm, offset=total_sleep_time_addr)

        # After Kernel 3.3 wall_to_monotonic is stored inside the timekeeper.
        timekeeper_addr = self.obj_profile.get_constant("timekeeper")
        if timekeeper_addr:
            return  self.obj_profile.timekeeper(
                vm=self.obj_vm, offset=timekeeper_addr).total_sleep_time

        # Just return an empty timespec.
        return self.obj_profile.timespec()

    def __add__(self, other):
        """Properly normalize this object from sec and nsec.

        based on set_normalized_timespec function.
        """
        if not isinstance(other, self.__class__):
            raise TypeError("Can only add timespec to timespec")

        sec = other.tv_sec + self.tv_sec
        nsec = other.tv_nsec + self.tv_nsec

        sec += nsec / self.NSEC_PER_SEC
        nsec = nsec % self.NSEC_PER_SEC

        result = self.obj_profile.timespec()
        result.tv_sec = sec
        result.tv_nsec = nsec

        return result

    def getboottime(self):
        result = self.wall_to_monotonic + self.total_sleep_time
        return result.tv_sec + result.tv_nsec / self.NSEC_PER_SEC

    def as_timestamp(self):
        secs = self.tv_sec + self.tv_nsec / self.NSEC_PER_SEC
        return self.obj_profile.UnixTimeStamp(value=secs)


class PermissionFlags(basic.Flags):
    """A Flags object for printing vm_area_struct permissions
    in a format like rwx or r-x"""

    def __str__(self):
        result = []
        value = self.v()
        for k, v in sorted(self.maskmap.items()):
            if value & v:
                result.append(k)
            else:
                result.append("-")

        return ''.join(result)

    def is_flag(self, flag):
        return self.v() & (1 << self.bitmap[flag])

    def is_executable(self):
        return self.is_flag('x')

    def is_readable(self):
        return self.is_flag('r')

    def is_writable(self):
        return self.is_flag('w')


class Linux32(basic.Profile32Bits, basic.BasicWindowsClasses):
    """A Linux profile which works with dwarfdump output files."""
    _md_os = "linux"
    _md_memory_model = "32bit"
    _md_type = "Kernel"

    def __init__(self, **kwargs):
        super(Linux32, self).__init__(**kwargs)
        self.add_classes(dict(
                list_head=list_head, hlist_head=hlist_head,
                dentry=dentry,
                task_struct=task_struct,
                timespec=timespec, inet_sock=inet_sock,
                PermissionFlags=PermissionFlags,
                ))
        self.add_overlay(linux_overlay)
        self.add_constants(default_text_encoding="utf8")

class Linux64(basic.ProfileLP64, Linux32):
    """Support for 64 bit linux systems."""



# Rekall Memory Forensics
# Copyright (C) 2012 Michael Cohen <scudette@gmail.com>
# Copyright (c) 2008 Brendan Dolan-Gavitt <bdolangavitt@wesleyan.edu>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
@author:       Michael Cohen <scudette@gmail.com>
@author:       AAron Walters and Brendan Dolan-Gavitt
@license:      GNU General Public License 2.0 or later
@contact:      awalters@volatilesystems.com,bdolangavitt@wesleyan.edu
@organization: Volatile Systems
"""
import re
import os

from rekall import args
from rekall import utils
from rekall.plugins import core
from rekall.plugins.windows import common
from rekall.plugins.windows.registry import registry


class PrintKey(common.WindowsCommandPlugin):
    "Print a registry key, and its subkeys and values"
    # Declare meta information associated with this plugin

    __name = "printkey"

    meta_info = {}
    meta_info['author'] = 'Brendan Dolan-Gavitt'
    meta_info['copyright'] = 'Copyright (c) 2007,2008 Brendan Dolan-Gavitt'
    meta_info['contact'] = 'bdolangavitt@wesleyan.edu'
    meta_info['license'] = 'GNU General Public License 2.0 or later'
    meta_info['url'] = 'http://moyix.blogspot.com/'
    meta_info['os'] = 'WIN_32_XP_SP2'
    meta_info['version'] = '1.0'

    @classmethod
    def args(cls, parser):
        """Declare the command line args we need."""
        super(PrintKey, cls).args(parser)

        parser.add_argument("-k", "--key", default="",
                            help="Registry key to print.")

        parser.add_argument("-o", "--hive_offsets", default=None,
                            action=args.ArrayIntParser, nargs="+",
                            help = 'Hive offsets to search (virtual)')

        parser.add_argument("-r", "--recursive", default=False,
                            action="store_true",
                            help = 'If set print the entire subtree.')


    def __init__(self, hive_offsets=None, key="", recursive=False, **kwargs):
        """Print all keys and values contained by a registry key.

        Args:
          hive_offsets: A list of hive offsets as found by hivelist (virtual
            address). If not provided we call hivescan ourselves and list the
            key on all hives.

          key: The key name to list. If not provided we list the root key in the
            hive.

          recursive: If set print the entire subtree.
        """
        super(PrintKey, self).__init__(**kwargs)
        self.profile = registry.RekallRegisteryImplementation(self.profile)
        self.hive_offsets = hive_offsets
        self.key = key
        self.recursive = recursive

    def _list_keys(self, reg, key=None):
        yield reg, key

        if self.recursive:
            for subkey in key.subkeys():
                for reg, subkey in self._list_keys(reg, subkey):
                    yield reg, subkey

    def list_keys(self):
        """Return the keys that match."""
        seen = set()
        if not self.hive_offsets:
            self.hive_offsets = list(self.get_plugin("hivescan").list_hives())

        for hive_offset in self.hive_offsets:
            if hive_offset in seen:
                continue

            seen.add(hive_offset)

            reg = registry.RegistryHive(
                profile=self.profile, kernel_address_space=self.kernel_address_space,
                hive_offset=hive_offset)

            key = reg.open_key(self.key)
            for reg, subkey in self._list_keys(reg, key):
                yield reg, subkey

    def voltext(self, key):
        """Returns a string representing (S)table or (V)olatile keys."""
        return "(V)" if key.obj_offset & 0x80000000 else "(S)"

    def render(self, renderer):
        renderer.format("Legend: (S) = Stable   (V) = Volatile\n\n")
        for reg, key in self.list_keys():
            if key:
                renderer.format("----------------------------\n")
                renderer.format("Registry: {0}\n", reg.Name)
                renderer.format("Key name: {0} {1:3s}\n", key.Name,
                                self.voltext(key))

                renderer.format("Last updated: {0}\n", key.LastWriteTime)
                renderer.format("\n")
                renderer.format("Subkeys:\n")

                for subkey in key.subkeys():
                    if not subkey.Name:
                        renderer.format("  Unknown subkey: {0}\n", subkey.Name.reason)
                    else:
                        renderer.format(u"  {1:3s} {0}\n",
                                        subkey.Name, self.voltext(subkey))

                renderer.format("\n")
                renderer.format("Values:\n")
                for value in key.values():
                    if value.Type == 'REG_BINARY':
                        data = value.DecodedData
                        if isinstance(data, basestring):
                            utils.WriteHexdump(renderer, value.DecodedData)
                    else:
                        renderer.format(u"{0:13} {1:15} : {3:3s} {2}\n",
                                        value.Type, value.Name, value.DecodedData,
                                        self.voltext(value))


class RegDump(core.DirectoryDumperMixin, common.WindowsCommandPlugin):
    """Dump all registry hives into a dump directory."""

    __name = 'regdump'

    @classmethod
    def args(cls, parser):
        """Declare the command line args we need."""
        super(RegDump, cls).args(parser)
        parser.add_argument("-o", "--hive_offsets", action=args.ArrayIntParser,
                            nargs="+", help='Hive offsets to search (virtual)')


    def __init__(self, hive_offsets=None, **kwargs):
        """Dump the registry from memory.

        Args:
          hive_offset: A list of hive offsets as found by hivelist (virtual
            address). If not provided we call hivescan ourselves and dump all
            hives found.

          dump_dir: Directory in which to dump hive files.
        """
        super(RegDump, self).__init__(**kwargs)
        self.hive_offsets = hive_offsets

    def dump_hive(self, hive_offset=None, reg=None, fd=None):
        """Write the hive into the fd.

        Args:
          hive_offset: The virtual offset where the hive is located.
          reg: Optionally an instance of registry.Registry helper. If provided
            hive_offset is ignored.
          fd: The file like object we write to.
        """
        if reg is None:
            reg = registry.RegistryHive(
                profile=self.profile, kernel_address_space=self.kernel_address_space,
                hive_offset=hive_offset)

        count = 0
        for data in reg.address_space.save():
            fd.write(data)
            count += len(data)
            self.session.report_progress("Dumping {0}Mb".format(count/1024/1024))

    def render(self, renderer):
        # Get all the offsets if needed.
        if not self.hive_offsets:
            self.hive_offsets = list(self.get_plugin("hivescan").list_hives())

        seen = set()
        for hive_offset in self.hive_offsets:
            if hive_offset in seen:
                continue

            reg = registry.RegistryHive(
                profile=self.profile, kernel_address_space=self.kernel_address_space,
                hive_offset=hive_offset)

            # Make up a filename for it, should be similar to the hive name.
            filename = reg.Name.rsplit("\\", 1).pop()

            # Sanitize it.
            filename = re.sub("[^a-zA-Z0-9_\-@ ]", "_", filename)

            # Make up the path.
            path = os.path.join(self.dump_dir, filename)

            renderer.section()
            renderer.format("Dumping {0} into \"{1}\"\n", reg.Name, path)

            with open(path, "wb") as fd:
                self.dump_hive(reg=reg, fd=fd)
                renderer.format("Dumped {0} bytes\n", fd.tell())



class HiveDump(registry.RegistryPlugin):
    """Prints out a hive"""

    __name = "hivedump"

    def _key_iterator(self, key):
        for subkey in key.subkeys():
            yield subkey
            for subsubkey in self._key_iterator(subkey):
                yield subsubkey

    def render(self, renderer):
        seen = set()

        for hive_offset in self.hive_offsets:
            if hive_offset in seen: continue

            reg = registry.RegistryHive(
                hive_offset=hive_offset,
                kernel_address_space=self.kernel_address_space,
                profile=self.profile)

            renderer.section()
            renderer.format("Hive {0}\n\n", reg.Name)

            renderer.table_header([("Last Written", "timestamp", "<24"),
                                   ("Key", "key", "")])

            for key in self._key_iterator(reg.root):
                renderer.table_row(key.LastWriteTime, key.Path)



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

__author__ = "Michael Cohen <scudette@google.com>"


from rekall.plugins.darwin import common


class DarwinLsof(common.DarwinProcessFilter):
    """List open files for processes."""

    __name = "lsof"

    def Lsof(self, proc):
        for i, fileproc in enumerate(proc.p_fd.fd_ofiles):
            # When the type of the glob is VNODE it contains a vnode struct.
            if fileproc.f_fglob.fg_type == "DTYPE_VNODE":
                vnode = fileproc.f_fglob.fg_data.dereference_as("vnode")
                yield fileproc, i, vnode.full_path

    def render(self, renderer):
        renderer.table_header([("PID", "pid", "8"),
                               ("Command", "command", "16"),
                               ("File Desc", "desc", "10"),
                               ("Path", "path", "20")])

        for proc in sorted(self.filter_processes(), key=lambda x: x.p_pid):
            for _, fd, path in self.Lsof(proc):
                renderer.table_row(proc.p_pid, proc.p_comm, fd, path)



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

# References:
# http://www.codemachine.com/article_kernelstruct.html#MMPFN

# pylint: disable=protected-access

from rekall import args
from rekall import testlib
from rekall import obj
from rekall import plugin
from rekall.plugins.windows import common
from rekall.plugins.overlays import basic

class ValueEnumeration(basic.Enumeration):
    """An enumeration which receives its value from a callable."""

    def __init__(self, choices=None, value=None, parent=None, **kwargs):
        super(ValueEnumeration, self).__init__(parent=parent, **kwargs)
        self.choices = choices
        if callable(value):
            value = value(parent)

        self.value = value

    def v(self, vm=None):
        return self.value


class PFNModification(obj.ProfileModification):
    """Installs types specific to the PFN database."""

    @classmethod
    def modify(cls, profile):
        # Some shortcuts to the most important information.
        profile.add_overlay({
                '_MMPTE': [None, {
                        'Valid': lambda x: x.u.Hard.Valid,
                        'PFN': lambda x: x.u.Hard.PageFrameNumber,
                        }],
                '_MMPFN': [None, {
                        "Type": [0, ["ValueEnumeration", dict(
                                    value=lambda x: x.u3.e1.PageLocation,
                                    choices={
                                        0: 'ZeroedPageList',
                                        1: 'FreePageList',
                                        2: 'StandbyPageList',
                                        3: 'ModifiedPageList',
                                        4: 'ModifiedNoWritePageList',
                                        5: 'BadPageList',
                                        6: 'ActiveAndValid',
                                        7: 'TransitionPage'
                                        }
                                    )]],
                        }],
                '_KDDEBUGGER_DATA64': [None, {
                        # This is the pointer to the PFN database.
                        'MmPfnDatabase': [None, ['Pointer', dict(
                                    target="Pointer",
                                    target_args=dict(
                                        target="Array",
                                        target_args=dict(target="_MMPFN"),
                                        ))]],
                        }],
                })
        profile.add_classes({
                "ValueEnumeration": ValueEnumeration,
                })


class VtoP(plugin.KernelASMixin, plugin.ProfileCommand):
    """Prints information about the virtual to physical translation."""

    __name = "vtop"

    PAGE_SIZE = 0x1000

    @classmethod
    def args(cls, parser):
        super(VtoP, cls).args(parser)
        parser.add_argument("virtual_address", action=args.IntParser,
                            help="Specify to see all the fops, even if they "
                            "are known.")

    def __init__(self, virtual_address=None, address_space=None, **kwargs):
        """Prints information about the virtual to physical translation.

        This is similar to windbg's !vtop extension.

        Args:
          virtual_address: The virtual address to describe.
          address_space: The address space to use (default the
            kernel_address_space).
        """
        super(VtoP, self).__init__(**kwargs)
        self.address_space = address_space or self.kernel_address_space
        self.address = virtual_address

    def _vtop_32bit(self, vaddr, address_space):
        """An implementation specific to the 32 bit intel address space."""
        pde_addr = ((address_space.dtb & 0xfffff000) |
                    ((vaddr & 0xffc00000) >> 20))

        pde_value = address_space.read_long_phys(pde_addr)
        yield "pde", pde_value, pde_addr

        if not address_space.entry_present(pde_value):
            yield "Invalid PDE", None, None
            return

        if address_space.page_size_flag(pde_value):
            yield "Large page mapped", address_space.get_four_meg_paddr(
                vaddr, pde_value), None
            return

        pte_addr = (pde_value & 0xfffff000) | ((vaddr & 0x3ff000) >> 10)
        pte_value = address_space.read_long_phys(pte_addr)
        yield "pte", pte_value, pte_addr

        if not address_space.entry_present(pde_value):
            yield "Invalid PTE", None, None
            return

        yield ("PTE mapped",
               address_space.get_phys_addr(vaddr, pte_value),
               pte_addr)

    def _vtop_32bit_pae(self, vaddr, address_space):
        """An implementation specific to the 32 bit PAE intel AS."""
        pdpte_addr = ((address_space.dtb & 0xfffffff0) |
                      ((vaddr & 0x7FC0000000) >> 27))

        pdpte_value = address_space._read_long_long_phys(pdpte_addr)
        yield "pdpte", pdpte_value, pdpte_addr

        if not address_space.entry_present(pdpte_value):
            yield "Invalid PDPTE", None, None
            return

        pde_addr = (pdpte_value & 0xfffff000) | ((vaddr & 0x3fe00000) >> 18)
        pde_value = address_space.read_long_phys(pde_addr)
        yield "pde", pde_value, pde_addr

        if not address_space.entry_present(pde_value):
            yield "Invalid PDE", None, None
            return

        if address_space.page_size_flag(pde_value):
            yield "Large page mapped", address_space.get_four_meg_paddr(
                vaddr, pde_value), None
            return

        pte_addr = (pde_value & 0xfffff000) | ((vaddr & 0x1ff000) >> 9)
        pte_value = address_space.read_long_phys(pte_addr)
        yield "pte", pte_value, pte_addr

        if not address_space.entry_present(pde_value):
            yield "Invalid PTE", None, None
            return

        yield ("PTE mapped",
               address_space.get_phys_addr(vaddr, pte_value),
               pte_addr)

    def _vtop_64bit(self, vaddr, address_space):
        """An implementation specific to the 64 bit intel address space."""
        pml4e_addr = ((address_space.dtb & 0xffffffffff000) |
                      ((vaddr & 0xff8000000000) >> 36))

        pml4e_value = address_space._read_long_long_phys(pml4e_addr)
        yield "pml4e", pml4e_value, pml4e_addr

        if not address_space.entry_present(pml4e_value):
            yield "Invalid PDE", None, None
            return

        pdpte_addr = ((pml4e_value & 0xffffffffff000) |
                      ((vaddr & 0x7FC0000000) >> 27))

        pdpte_value = address_space._read_long_long_phys(pdpte_addr)
        yield "pdpte", pdpte_value, pdpte_addr

        if address_space.page_size_flag(pdpte_value):
            yield "One Gig page", address_space.get_one_gig_paddr(
                vaddr, pdpte_value), None
            return

        pde_addr = ((pdpte_value & 0xffffffffff000) |
                    ((vaddr & 0x3fe00000) >> 18))
        pde_value = address_space.read_long_phys(pde_addr)
        yield "pde", pde_value, pde_addr

        if not address_space.entry_present(pde_value):
            yield "Invalid PDE", None, None
            return

        if address_space.page_size_flag(pde_value):
            yield "Large page mapped", address_space.get_four_meg_paddr(
                vaddr, pde_value), None
            return

        pte_addr = (pde_value & 0xffffffffff000) | ((vaddr & 0x1ff000) >> 9)
        pte_value = address_space.read_long_phys(pte_addr)
        yield "pte", pte_value, pte_addr

        if not address_space.entry_present(pte_value):
            yield "Invalid PTE", None, None
            return

        yield ("PTE mapped",
               address_space.get_phys_addr(vaddr, pte_value),
               pte_addr)

    def vtop(self, virtual_address, address_space=None):
        """Translate the virtual_address using the address_space."""
        if address_space is None:
            address_space = self.kernel_address_space

        if address_space.metadata("memory_model") == "64bit":
            function = self._vtop_64bit
        else:
            if address_space.metadata("pae"):
                function = self._vtop_32bit_pae
            else:
                function = self._vtop_32bit

        return function(virtual_address, address_space)

    def render(self, renderer):
        if self.address is None:
            return

        renderer.format("Virtual {0:#08x} Page Directory 0x{1:08x}\n",
                        self.address, self.address_space.dtb)

        for name, value, address in self.vtop(self.address, self.address_space):
            if address:
                renderer.format("{0}@ {2:#08x} = {1:#08x}\n",
                                name, value, address)
            elif value:
                renderer.format("{0} {1:#08x}\n", name, value)
            else:
                renderer.format("{0}\n", name)

        physical_address = self.address_space.vtop(self.address)
        if physical_address is None:
            renderer.format("Physical Address Invalid\n")
        else:
            renderer.format("Physical Address {0:#08x}\n", physical_address)


class PFNInfo(common.WindowsCommandPlugin):
    """Prints information about an address from the PFN database."""

    __name = "pfn"

    # Size of page.
    PAGE_SIZE = 0x1000
    PAGE_BITS = 12

    def __init__(self, pfn=None, physical_address=None, **kwargs):
        """Prints information about the physical PFN entry.

        Args:
          pfn: A page file number to display.
          physical_address: The physical address to print information about.
        """
        super(PFNInfo, self).__init__(**kwargs)

        self.profile = PFNModification(self.profile)

        # We prefer our own private version of the kdbg.
        self.kdbg = self.profile.Object("_KDDEBUGGER_DATA64", offset=self.kdbg,
                                        vm=self.kernel_address_space)

        # A reference to the pfn database.
        self.pfn_database = self.kdbg.MmPfnDatabase.dereference().dereference()
        self.pfn = pfn
        self.physical_address = physical_address

    def pfn_record(self, pfn=None, physical_address=None):
        """Returns the pfn record for a pfn or a virtual address."""
        if physical_address is not None:
            pfn = int(physical_address) / self.PAGE_SIZE

        if pfn is None:
            raise RuntimeError("PFN not provided.")

        # Return the pfn record.
        return self.pfn_database[pfn]

    def render(self, renderer):
        pfn = self.pfn
        if self.physical_address is not None:
            pfn = int(self.physical_address) / self.PAGE_SIZE

        pfn_obj = self.pfn_record(pfn)

        renderer.format("""    PFN 0x{0:08X} at kernel address 0x{1:016X}
""", pfn, pfn_obj.obj_offset)

        # The flags we are going to print.
        flags = {"M": "Modified",
                 "P": "ParityError",
                 "R": "ReadInProgress",
                 "W": "WriteInProgress"}

        short_flags_string = "".join(
            [k for k, v in flags.items() if pfn_obj.u3.e1.m(v) == 0])

        long_flags_string = " ".join(
            [v for k, v in flags.items() if pfn_obj.u3.e1.m(v) == 0])

        containing_page = int(pfn_obj.u4.PteFrame)
        pte_physical_address = ((containing_page << self.PAGE_BITS) |
                                (int(pfn_obj.PteAddress) & 0xFFF))

        renderer.format("""    flink       {0:08X}  blink / share count {1:016X}
    pteaddress (VAS) 0x{2:016X}  (Phys AS) 0x{3:016X}
    reference count {4:04X}   color {5}
    containing page        0x{6:08X}  {7}     {8}
    {9}
    """, pfn_obj.u1.Flink, pfn_obj.u2.Blink,
                        pfn_obj.PteAddress,
                        pte_physical_address,
                        pfn_obj.u3.e2.ReferenceCount,
                        pfn_obj.u3.e1.m("PageColor") or pfn_obj.u4.PageColor,
                        containing_page,
                        pfn_obj.Type,
                        short_flags_string,
                        long_flags_string)


class PTE(common.WindowsCommandPlugin):
    """Prints information about a PTE."""

    __name = "pte"


    def __init__(self, virtual_address=None, pte_address=None, **kwargs):
        """Prints information about a PTE.

        Similar to windbg's !pte extension.

        Args:
          virtual_address: The virtual address to describe.
          pte_address: An address of a PTE record.
        """
        super(PTE, self).__init__(**kwargs)
        self.vtop = VtoP(session=self.session)
        self.pte_address = pte_address
        self.virtual_address = virtual_address


    def render(self, renderer):
        if self.virtual_address is not None:
            for name, _, _ in self.vtop.vtop(
                self.virtual_address, self.kernel_address_space):
                if name == "pte":
                    break


class PtoV(common.WinProcessFilter):
    """Converts a physical address to a virtual address."""

    __name = "ptov"

    PAGE_SIZE = 0x1000
    PAGE_BITS = 12

    def __init__(self, physical_address=None, **kwargs):
        """Converts a physical address to a virtual address."""
        super(PtoV, self).__init__(**kwargs)

        # Get a handle to the pfninfo plugin
        self.pfn_plugin = self.session.plugins.pfn(session=self.session)
        self.physical_address = physical_address

    def _ptov_x86(self, physical_address):
        """An implementation of ptov for x86."""
        result = physical_address & 0xFFF

        # Get the pte for this physical_address using the pfn database.
        pfn_obj = self.pfn_plugin.pfn_record(physical_address >> self.PAGE_BITS)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PTE invalid."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pte_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pte_address << 10) & 0x3FF000

        # Get the PDE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PDE invalid (Is this a large page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pde_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pde_address << 20) & 0xffc00000

        # Now get the DTB.
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        containing_page = int(pfn_obj.u4.PteFrame)
        dtb_address = containing_page << self.PAGE_BITS

        return result, (("DTB", dtb_address),
                        ("PDE", pde_address),
                        ("PTE", pte_address))

    def _ptov_x86_pae(self, physical_address):
        """An implementation of ptov for x86 pae."""
        result = physical_address & 0xFFF
        # Get the pte for this physical_address using the pfn database.
        pfn_obj = self.pfn_plugin.pfn_record(physical_address >> self.PAGE_BITS)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PTE invalid."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pte_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pte_address << 9) & 0x1FF000

        # Get the PDE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PDE invalid (Is this a large page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pde_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pde_address << 18) & 0x3fe00000

        # Get the PDPTE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject(
                "PDPTE invalid (Is this a one gig page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pdpte_address = ((containing_page << self.PAGE_BITS) |
                         (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pdpte_address << 27) & 0x7FC0000000

        # Now get the DTB.
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        containing_page = int(pfn_obj.u4.PteFrame)
        dtb_address = containing_page << self.PAGE_BITS

        return result, (("DTB", dtb_address),
                        ("PDPTE", pdpte_address),
                        ("PDE", pde_address),
                        ("PTE", pte_address))

    def _ptov_x64(self, physical_address):
        """An implementation of ptov for x64."""
        result = physical_address & 0xFFF

        # Get the pte for this physical_address using the pfn database.
        pfn_obj = self.pfn_plugin.pfn_record(physical_address >> self.PAGE_BITS)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PTE invalid."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pte_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pte_address << 9) & 0x1FF000

        # Get the PDE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PDE invalid (Is this a large page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pde_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pde_address << 18) & 0x3fe00000

        # Get the PDPTE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject(
                "PDPTE invalid (Is this a one gig page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pdpte_address = ((containing_page << self.PAGE_BITS) |
                         (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pdpte_address << 27) & 0x7FC0000000

        # Get the PML4E now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PML4E invalid."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pml4e_address = ((containing_page << self.PAGE_BITS) |
                         (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pml4e_address << 36) & 0xff8000000000

        # Now get the DTB.
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        containing_page = int(pfn_obj.u4.PteFrame)
        dtb_address = containing_page << self.PAGE_BITS

        return result, (("DTB", dtb_address),
                        ("PML4E", pml4e_address),
                        ("PDPTE", pdpte_address),
                        ("PDE", pde_address),
                        ("PTE", pte_address))

    def ptov(self, physical_address):
        """Convert the physical address to a virtual address.

        Returns:
          a tuple (_EPROCESS of owning process, virtual address in process AS).
        """
        if self.kernel_address_space.metadata("memory_model") == "32bit":
            if self.kernel_address_space.metadata("pae"):
                return self._ptov_x86_pae(physical_address)
            else:
                return self._ptov_x86(physical_address)
        elif self.kernel_address_space.metadata("memory_model") == "64bit":
            return self._ptov_x64(physical_address)

        return obj.NoneObject("Memory model not supported."), []

    def render(self, renderer):
        if self.physical_address is None:
            return

        result, structures = self.ptov(self.physical_address)
        if result:
            renderer.format("Physical Address {0:#x} => "
                            "Virtual Address {1:#x}\n",
                            self.physical_address, result)

            for type, phys_addr in structures:
                renderer.format("{0} @ {1:#x}\n", type, phys_addr)
        else:
            renderer.format("Error converting Physical Address {0:#x}: "
                            "{1!r}\n", self.physical_address, result)


class DTBScan(common.WinProcessFilter):
    """Scans the physical memory for DTB values.

    This plugin can compare the DTBs found against the list of known processes
    to find hidden processes.
    """

    __name = "dtbscan"

    @classmethod
    def args(cls, parser):
        super(DTBScan, cls).args(parser)
        parser.add_argument("--limit", action=args.IntParser, default=0,
                            help="Stop scanning after this many mb.")

    def __init__(self, limit=None, **kwargs):
        super(DTBScan, self).__init__(**kwargs)
        self.limit = limit

    def render(self, renderer):
        ptov = self.session.plugins.ptov(session=self.session)
        pslist = self.session.plugins.pslist(session=self.session)
        pfn_plugin = self.session.plugins.pfn(session=self.session)

        # Known tasks:
        known_tasks = set()
        for task in pslist.list_eprocess():
            known_tasks.add(task.obj_offset)

        renderer.table_header([("DTB", "dtb", "[addrpad]"),
                               ("VAddr", "vaddr", "[addrpad]"),
                               ("_EPROCESS", "task", "[addrpad]"),
                               ("Image Name", "filename", "<20"),
                               ("Known", "known", "")])

        seen_dtbs = set()

        # Now scan all the physical address space for DTBs.
        for _ in self.physical_address_space.get_available_addresses():
            start, length = _
            for page in range(start, start + length, 0x1000):
                self.session.report_progress("Scanning 0x%08X (%smb)" % (
                        page, page/1024/1024))

                # Quit early if requested to.
                if self.limit and page > self.limit:
                    return

                virtual_address, results = ptov.ptov(page)
                if virtual_address:
                    dtb = results[0][1]
                    if dtb not in seen_dtbs:
                        seen_dtbs.add(dtb)

                        # The _EPROCESS address is stored as the
                        # KernelStackOwner for the pfn of this dtb.
                        task = pfn_plugin.pfn_record(
                            dtb >> 12).u1.Flink.dereference_as(
                            "_EPROCESS")

                        if not task:
                            task = obj.NoneObject("Invalid")
                            filename = "Process not Found!"
                        else:
                            filename = task.ImageFileName

                        va, _ = ptov.ptov(dtb)
                        renderer.table_row(dtb, va, task, filename,
                                           task.obj_offset in known_tasks)

class TestDTBScan(testlib.SimpleTestCase):
    PARAMETERS = dict(
        commandline="dtbscan --limit 10mb",
        )
